"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# -*- coding: utf-8 -*-

import numpy as np

from qrisp import (
    PGate,
    QuantumBool,
    QuantumCircuit,
    QuantumSession,
    QuantumVariable,
    convert_to_qb_list,
    RYGate,
    TruthTable,
    XGate,
    bin_rep,
    cx,
    h,
    invert,
    mcx,
    s,
    s_dg,
    x,
)
from qrisp.misc.GMS_tools import GXX_wrapper


# Interface function to quickly change between different implementations of
# multi controlled not gates
def multi_cx(n, method=None):
    # from qrisp.circuit import transpile

    if method == "gms":
        return gms_multi_cx(n)

    elif method == "gray_pt":
        return pt_multi_cx(n)

    elif method == "gray_pt_inv":
        return pt_multi_cx(n).inverse()

    elif method in ["gray", "auto", None]:
        return gray_multi_cx(n)

    else:
        raise Exception('method "' + method + '" not implemented')


# Function to synthesize a multi controlled CX gate from GMS gates
def gms_multi_cx(n):
    qc = QuantumCircuit(n + 1)

    if n == 2:
        for i in range(3):
            qc.ry(np.pi / 2, i)

        qc.rz(np.pi / 4, 2)

        qc.append(GXX_wrapper(3, 3 * [3 * [np.pi / 2]]), qc.qubits)

        for i in range(3):
            qc.rx(-np.pi / 2, i)

        qc.rz(-np.pi / 2, 2)

        for i in range(3):
            qc.rx(-np.pi / 4, i)

        qc.append(GXX_wrapper(3, 3 * [3 * [np.pi / 4]]), qc.qubits)

        qc.rz(np.pi / 2, 2)

        qc.append(GXX_wrapper(3, 3 * [3 * [np.pi / 2]]), qc.qubits)

        for i in range(3):
            qc.rx(np.pi / 2, i)
            qc.ry(-np.pi / 2, i)

        result = qc.to_gate()

        result.name = "GMS toffoli"

        return result
    else:
        raise Exception(str(n) + "-controlled x gate not implemented for gms method")


def gray_pt_mcx(n, ctrl_state):
    input_qv = QuantumVariable(n)
    output_qv = QuantumVariable(1, qs=input_qv.qs)

    tt_str = 2 ** n * ["0"]
    tt_str[int(ctrl_state[::-1], 2)] = "1"

    tt = TruthTable([tt_str])

    tt.q_synth(input_qv, output_qv, method="gray_pt")

    res = input_qv.qs.copy()
    return res.to_gate(f"pt{n}cx")


# Function to synthesize a phase tolerant multi controlled X gate
def pt_multi_cx(n, reduced=False):
    res = QuantumCircuit(n + 1)

    if n == 1:
        res.cx(0, 1)

    # The special cases n = 2 and n = 3 are handled as described in
    # https://arxiv.org/pdf/1508.03273.pdf
    elif n == 2:
        if reduced:
            res.append(reduced_margolus_qc.to_gate("reduced margolus"), res.qubits)
        else:
            res.append(margolus_qc.to_gate("margolus"), res.qubits)

    elif n == 3:
        if reduced:
            res.append(reduced_maslov_qc.to_gate("reduced maslov"), res.qubits)
        else:
            res.append(maslov_qc.to_gate("maslov"), res.qubits)
    else:
        input_qv = QuantumVariable(n)
        output_qv = QuantumVariable(1, qs=input_qv.qs)

        tt = TruthTable([(2 ** (n) - 1) * "0" + "1"])

        tt.q_synth(input_qv, output_qv, method="gray_pt")

        res = input_qv.qs.copy()

    return res.to_gate(f"pt{n}cx")


# Toffoli Implementation according to https://arxiv.org/pdf/1206.0758.pdf
toffoli_qc = QuantumCircuit(3)

toffoli_qc.h(2)
toffoli_qc.p(-np.pi / 4, [0, 1])
toffoli_qc.cx(2, 0)
toffoli_qc.cx(1, 2)
toffoli_qc.p(np.pi / 4, 0)
toffoli_qc.cx(1, 0)
toffoli_qc.p(np.pi / 4, 2)
toffoli_qc.cx(1, 2)
toffoli_qc.p(-np.pi / 4, 0)
toffoli_qc.cx(2, 0)
toffoli_qc.p(np.pi / 4, 0)
toffoli_qc.p(-np.pi / 4, 2)
toffoli_qc.cx(1, 0)
toffoli_qc.h(2)
toffoli_qc.qubits[1], toffoli_qc.qubits[0] = toffoli_qc.qubits[0], toffoli_qc.qubits[1]

margolus_qc = QuantumCircuit(3)
G = RYGate(np.pi / 4)

margolus_qc.append(G, 2)
margolus_qc.cx(1, 2)
margolus_qc.append(G, 2)
margolus_qc.cx(0, 2)

reduced_margolus_qc = margolus_qc.copy()

margolus_qc.append(G.inverse(), 2)
margolus_qc.cx(1, 2)
margolus_qc.append(G.inverse(), 2)

maslov_qc = QuantumCircuit(4)

maslov_qc.h(3)
maslov_qc.t(3)
maslov_qc.cx(2, 3)
maslov_qc.t_dg(3)
maslov_qc.h(3)
maslov_qc.cx(0, 3)
maslov_qc.t(3)
maslov_qc.cx(1, 3)
maslov_qc.t_dg(3)
maslov_qc.cx(0, 3)

reduced_maslov_qc = maslov_qc.copy()

maslov_qc.t(3)
maslov_qc.cx(1, 3)
maslov_qc.t_dg(3)
maslov_qc.h(3)
maslov_qc.t(3)
maslov_qc.cx(2, 3)
maslov_qc.t_dg(3)
maslov_qc.h(3)


# Ancilla supported multi controlled X gates from https://arxiv.org/pdf/1508.03273.pdf
def maslov_mcx(n, ctrl_state=-1):
    if not isinstance(ctrl_state, str):
        if ctrl_state == -1:
            ctrl_state += 2 ** n
        ctrl_state = bin_rep(ctrl_state, n)

    res = QuantumCircuit(n + 1)
    for i in range(len(ctrl_state)):
        if ctrl_state[i] == "0":
            res.x(i)

    if n == 1:
        res.cx(0, 1)
    elif n == 2:
        res.append(toffoli_qc.to_gate("toffoli"), res.qubits)
    elif n == 3:
        res.add_qubit()
        res.append(margolus_qc.to_gate(), [0, 1, 3])
        res.append(toffoli_qc.to_gate(), [2, 3, 4])
        res.append(margolus_qc.inverse().to_gate(), [0, 1, 3])
    elif n == 4:
        res.add_qubit()
        res.append(maslov_qc.to_gate(), [0, 1, 2, 4])
        res.append(toffoli_qc.to_gate(), [3, 4, 5])
        res.append(maslov_qc.inverse().to_gate(), [0, 1, 2, 4])
    else:
        raise Exception('Multi CX for method "Maslov" only defined for n <= 4')

    for i in range(len(ctrl_state)):
        if ctrl_state[i] == "0":
            res.x(i)

    return res.to_gate(f"maslov {n}cx")


# Function to synthesize a multi controlled X gate using gray synthesis
def gray_multi_cx(n):
    qs = QuantumSession()
    input_qv = QuantumVariable(n, qs)
    output_qv = QuantumVariable(1, qs)

    if n == 1:
        qs.cx(0, 1)
    elif n == 2:
        qs = toffoli_qc.copy()

    else:
        tt_array = np.zeros((2 ** n, 1))
        tt_array[-1, 0] = 1

        tt = TruthTable(tt_array)

        tt.q_synth(input_qv, output_qv, method="gray")

    result = qs.to_gate()
    result.name = "gray multi cx"

    return result


# Ancilla supported multi controlled X with logarithmic depth based on
# https://www.iccs-meeting.org/archive/iccs2022/papers/133530169.pdf
def balauca_mcx(input_qubits, target, ctrl_state=None, phase=None):
    hybrid_mcx(
        input_qubits, target, ctrl_state=ctrl_state, phase=phase, num_ancilla=np.inf
    )


# Algorithm based on https://link.springer.com/article/10.1007/s10773-017-3389-4
def yong_mcx(input_qubits, target, ancilla=None, ctrl_state=None):
    if isinstance(target, list):
        if len(target) != 1:
            raise Exception("Tried to execute yong algorithm with multiple targets")
        target = target[0]

    if ctrl_state is not None:
        for i in range(len(input_qubits)):
            input_qubits[i].ctrl_state = ctrl_state[i]

    if len(input_qubits) < 3:
        ctrl_state = ""
        for i in range(len(input_qubits)):
            if hasattr(input_qubits[i], "ctrl_state"):
                ctrl_state += input_qubits[i].ctrl_state
            else:
                ctrl_state += "1"

    if len(input_qubits) == 2:
        if ancilla is None:
            mcx(input_qubits, target, method="gray", ctrl_state=ctrl_state)
        else:
            target.qs().append(
                gray_pt_mcx(2, ctrl_state=ctrl_state), input_qubits + [target]
            )
            # mcx(input_qubits, target, method = "gray_pt", ctrl_state = ctrl_state)
        return

    ancilla_allocated = False

    if ancilla is None:
        ancilla_allocated = True
        ancilla_bl = QuantumBool(name="yong_anc*")
        ancilla = ancilla_bl[0]

    n = len(input_qubits)

    partition_k_1 = input_qubits[n // 2:]
    partition_k_2 = input_qubits[: n // 2]

    h(target)

    yong_mcx(input_qubits=partition_k_1, target=ancilla, ancilla=partition_k_2[-1])

    s(ancilla)

    yong_mcx(
        input_qubits=partition_k_2 + [target], target=ancilla, ancilla=partition_k_1[-1]
    )

    s_dg(ancilla)

    with invert():
        s(ancilla)

        yong_mcx(
            input_qubits=partition_k_2 + [target],
            target=ancilla,
            ancilla=partition_k_1[0],
        )

        s_dg(ancilla)

        yong_mcx(input_qubits=partition_k_1, target=ancilla, ancilla=partition_k_2[0])

    h(target)

    if ancilla_allocated:
        ancilla_bl.delete()

    if ctrl_state is not None:
        for i in range(len(input_qubits)):
            del input_qubits[i].ctrl_state


# Hybrid algorithm of yong and balauca with customizable ancilla qubit count.
# Performs several balauca layers and cancels the recursion with yong.
def hybrid_mcx(
        input_qubits,
        target,
        ctrl_state=None,
        phase=None,
        num_ancilla=np.inf,
        num_dirty_ancilla=0,
):
    """
    Function to dynamically generate mcx gates for a given amount of ancilla qubits.

    Parameters
    ----------
    input_qubits : list[Qubit]
        The Qubits to control on.
    target : Qubit
        The Qubit to target.
    ctrl_state : str, optional
        The control state to activate the X Gate on. The default is "11...".
    phase : float or sympy.Symbol, optional
        If given, this function performs a mcp instead of a mcx. The default is None.
    num_ancilla : int, optional
        The amount of ancillae this function is allowed to use. The default is np.inf.
    count_yong : bool, optional
        If set to False, the ancilla that is used by the yong recursion termination is
        not counted (because it can be dirty). The default is True.

    Returns
    -------
    None.

    """

    input_qubits = list(input_qubits)
    for i in range(len(input_qubits)):
        if isinstance(input_qubits[i], QuantumBool):
            input_qubits[i] = input_qubits[i][0]

    if ctrl_state is not None:
        for i in range(len(input_qubits)):
            input_qubits[i].ctrl_state = ctrl_state[i]

    target = list(target)

    qs = target[0].qs()

    if len(input_qubits) <= 3 or num_ancilla == 0:
        ctrl_state = ""
        for qb in input_qubits:
            if hasattr(qb, "ctrl_state"):
                ctrl_state += qb.ctrl_state
            else:
                ctrl_state += "1"

        if len(input_qubits) == 2:
            if phase is None:
                qs.append(
                    XGate().control(2, ctrl_state=ctrl_state, method="gray"),
                    input_qubits + [target],
                )
            else:
                qs.append(
                    PGate(phase).control(2, ctrl_state=ctrl_state),
                    input_qubits + [target],
                )

        elif num_dirty_ancilla and phase is None:
            balauca_dirty(
                input_qubits, target, k=num_dirty_ancilla, ctrl_state=ctrl_state
            )

        else:
            if phase is None:
                qs.append(
                    XGate().control(
                        len(input_qubits), ctrl_state=ctrl_state, method="gray"
                    ),
                    input_qubits + [target],
                )
            else:
                qs.append(
                    PGate(phase).control(len(input_qubits), ctrl_state=ctrl_state),
                    input_qubits + [target],
                )
        return

    structure = structure_decider(len(input_qubits), num_ancilla)  # [::-1]
    layer_input = []
    layer_structure = []
    layer_output = []
    remainder = list(input_qubits)

    for i in range(len(structure)):
        if not structure[0] <= len(remainder):
            break

        layer_output.append(QuantumBool(name="balauca_anc*"))

        for j in range(structure[0]):
            layer_input.append(remainder.pop(0))

        layer_structure.append(structure.pop(0))

    balauca_layer(layer_input, layer_output, structure=layer_structure, invert=False)

    hybrid_mcx(
        layer_output + remainder,
        target,
        num_ancilla=num_ancilla - len(layer_output),
        num_dirty_ancilla=num_dirty_ancilla,
        phase=phase,
    )

    balauca_layer(layer_input, layer_output, structure=layer_structure, invert=True)

    [qbl.delete() for qbl in layer_output]

    if ctrl_state is not None:
        for i in range(len(input_qubits)):
            del input_qubits[i].ctrl_state

    return


def balauca_layer(input_qubits, output_qubits, structure, invert=False):
    if not output_qubits:
        return

    qs = output_qubits[0].qs()
    input_qubits = list(input_qubits)

    counter = 0
    for i in range(len(output_qubits)):
        if structure[i] == 3:
            ctrl_qubits = [
                input_qubits[counter],
                input_qubits[counter + 1],
                input_qubits[counter + 2],
            ]
            counter += 3

            ctrl_state = ""
            for qb in ctrl_qubits:
                if hasattr(qb, "ctrl_state"):
                    ctrl_state += qb.ctrl_state
                else:
                    ctrl_state += "1"

            gate = XGate().control(3, method="gray_pt", ctrl_state=ctrl_state)

            if invert:
                gate = gate.inverse()

            qs.append(gate, ctrl_qubits + [output_qubits[i]])
        else:
            ctrl_qubits = [input_qubits[counter], input_qubits[counter + 1]]
            counter += 2

            ctrl_state = ""
            for qb in ctrl_qubits:
                if hasattr(qb, "ctrl_state"):
                    ctrl_state += qb.ctrl_state
                else:
                    ctrl_state += "1"

            gate = XGate().control(2, method="gray_pt", ctrl_state=ctrl_state)

            if invert:
                gate = gate.inverse()

            qs.append(gate, ctrl_qubits + [output_qubits[i]])

    return


def structure_decider(n, k):
    # Each element of a Balauca layer reduces the amount of
    # Control qubits for the next layer either by 1 (margolous gate)
    # or by 2 (phase toleratn maslov gate). Ideally, we reduce the amount
    # of control qubits to 3, so we can cancel the recursion
    # and deploy a Toffoli.

    # If we have n controls, p = n - 2 is the amount of reductions,
    # that need to be performed.

    # If we have k ancillae at our disposal, we need to satisfy:

    # p = 2*triple_mcx + (k-triple_mcx_count) = k + triple_mcx_count

    # <=> triple_mcx_count = n - 2 - k

    triple_mcx_count = n - 2 - k

    if triple_mcx_count <= 0:
        if n % 2:
            return (n - 3) // 2 * [2] + [3]
        else:
            return n // 2 * [2]
    elif triple_mcx_count > k:
        return k * [3]
    else:
        return triple_mcx_count * [3] + (k - triple_mcx_count) * [2]


def vchain_2_dirty(control, target, dirty_ancillae=None):
    control = convert_to_qb_list(control)
    target = convert_to_qb_list(target)

    if len(control) == 1:
        cx(control, target)
        return
    elif len(control) == 2:
        mcx(control, target, method="gray")

    n = len(control)
    k = n - 2

    dirty_ancilla_qbls = []
    if dirty_ancillae is None:
        dirty_ancilla_qbls = [QuantumBool(name="vchain_2_dirty*") for i in range(k)]
        dirty_ancillae = [qbl[0] for qbl in dirty_ancilla_qbls]

    def reduced_margolus(control, target):
        qs = target.qs()

        control = list(control)
        qs.append(reduced_margolus_qc.to_gate("reduced_margolus"), control + [target])
        # qs.append(XGate().control(2), control + [target])

    def margolus(control, target):
        qs = target.qs()

        control = list(control)
        qs.append(margolus_qc.to_gate("margolus"), control + [target])
        # qs.append(XGate().control(2), control + [target])

    control_temp_list = list(control)
    dirty_ancillae_temp_list = list(dirty_ancillae)
    qubit_list = []

    qubit_list.append(control_temp_list.pop(0))
    qubit_list.append(control_temp_list.pop(0))

    while control_temp_list:
        qubit_list.append(dirty_ancillae_temp_list.pop(0))
        qubit_list.append(control_temp_list.pop(0))

    qubit_list.append(target[0])

    m = len(qubit_list)

    for i in range(k):
        if i == k - 1:
            margolus(
                [qubit_list[m - 5 - 2 * i], qubit_list[m - 4 - 2 * i]],
                qubit_list[m - 3 - 2 * i],
            )
        else:
            reduced_margolus(
                [qubit_list[m - 5 - 2 * i], qubit_list[m - 4 - 2 * i]],
                qubit_list[m - 3 - 2 * i],
            )

    for i in range(k):
        if i == k - 1:
            reduced_margolus(
                [qubit_list[2 * i + 2], qubit_list[2 * i + 3]], qubit_list[2 * i + 4]
            )
        else:
            with invert():
                reduced_margolus(
                    [qubit_list[2 * i + 2], qubit_list[2 * i + 3]],
                    qubit_list[2 * i + 4],
                )

    for i in range(k):
        if i == k - 1:
            with invert():
                margolus(
                    [qubit_list[m - 5 - 2 * i], qubit_list[m - 4 - 2 * i]],
                    qubit_list[m - 3 - 2 * i],
                )
        else:
            reduced_margolus(
                [qubit_list[m - 5 - 2 * i], qubit_list[m - 4 - 2 * i]],
                qubit_list[m - 3 - 2 * i],
            )

    for i in range(k):
        if i == k - 1:
            with invert():
                reduced_margolus(
                    [qubit_list[2 * i + 2], qubit_list[2 * i + 3]],
                    qubit_list[2 * i + 4],
                )
        else:
            with invert():
                reduced_margolus(
                    [qubit_list[2 * i + 2], qubit_list[2 * i + 3]],
                    qubit_list[2 * i + 4],
                )

    [qbl.delete() for qbl in dirty_ancilla_qbls]


def balauca_dirty(control, target, k, dirty_ancillae=None, ctrl_state=None):
    control = convert_to_qb_list(control)
    target = convert_to_qb_list(target)
    qs = target[0].qs()

    n = len(control)

    k = min((n - 2) // 2 + 1, k)

    if k <= 1:
        qs.append(
            XGate().control(n, ctrl_state=ctrl_state, method="gray"),
            list(control) + [target],
        )
        return

    if ctrl_state is not None:
        for i in range(len(ctrl_state)):
            if ctrl_state[i] == "0":
                x(control[i])

    def reduced_maslov(control, target):
        qs = target.qs()

        qs.append(reduced_maslov_qc.to_gate("reduced_maslov"), control + [target])
        # qs.append(XGate().control(2), control + [target])

    m_1 = int(np.ceil((n - 2 * (k - 1)) / 2))
    m_2 = int(np.floor((n - 2 * (k - 1)) / 2))

    dirty_ancilla_qbls = []
    if dirty_ancillae is None:
        dirty_ancilla_qbls = [
            QuantumBool(name="balauca_dirty*", qs=qs) for i in range(k)
        ]
        dirty_ancillae = [qbl[0] for qbl in dirty_ancilla_qbls]

    upper_block_qubits = control[:m_1] + [dirty_ancillae[0]]
    lower_block_qubits = control[-m_2:] + [dirty_ancillae[-1]]

    def balauca_dirty_helper(control, target, ancillae):
        if len(ancillae) == 0:
            vchain_2_dirty(control, target, dirty_ancillae=lower_block_qubits)
            # mcx(control, target)
            return

        reduced_maslov([ancillae[-1]] + control[-2:], target)

        balauca_dirty_helper(control[:-2], ancillae[-1], ancillae[:-1])

        with invert():
            reduced_maslov([ancillae[-1]] + control[-2:], target)

    vchain_2_dirty(lower_block_qubits, target, dirty_ancillae=upper_block_qubits)

    balauca_dirty_helper(control[:-m_2], dirty_ancillae[-1], dirty_ancillae[:-1])

    vchain_2_dirty(lower_block_qubits, target, dirty_ancillae=upper_block_qubits)

    balauca_dirty_helper(control[:-m_2], dirty_ancillae[-1], dirty_ancillae[:-1])

    [qbl.delete() for qbl in dirty_ancilla_qbls]

    if ctrl_state is not None:
        for i in range(len(ctrl_state)):
            if ctrl_state[i] == "0":
                x(control[i])



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp import conjugate, QuantumFloat, p, QFT, control, PTControlledOperation
import numpy as np

def test_iteration_env():
    
    def fourier_adder(qf, n):
    
        with conjugate(QFT)(qf):
    
            for i in range(qf.size):
                p(n*np.pi*2**(i-qf.size+1), qf[i])
                
    
    qf = QuantumFloat(5)
    fourier_adder(qf, 3)
    
    assert qf.get_measurement() == {3: 1.0}
    
    fourier_adder(qf, 2)
    
    assert qf.get_measurement() == {5: 1.0}
    
    ctrl = QuantumFloat(1)
    qf = QuantumFloat(5)

    with control(ctrl):
        fourier_adder(qf, 3)
        
    transpiled_qc = qf.qs.transpile(1)
    
    for instr in transpiled_qc.data:
        
        if "conjugator" in instr.op.name:
            assert not isinstance(instr.op, PTControlledOperation)
        
        
        
        


"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# Created by ann81984 at 07.05.2022
import time

from qrisp.core import QuantumSession
from qrisp.arithmetic import QuantumFloat
from numpy.linalg import norm
from qiskit import execute, Aer
from qrisp.interface import convert_circuit


def test_unitary_calculation():
    n = 2

    qs = QuantumSession()

    x = QuantumFloat(n, 0, qs, signed=True)
    y = QuantumFloat(n, 0, qs, signed=False)

    s = x * y
    # qs = s.qs
    qc = qs.compile(1)

    start = time.time()
    test_unitary_1 = qc.get_unitary()
    end = time.time()

    print("Qrisp calculation time = %s" % (end - start))

    ###################
    qc.qubits = qc.qubits[::-1]
    qiskit_qc = convert_circuit(qc, target_api="qiskit")

    backend = Aer.get_backend("unitary_simulator")

    start = time.time()
    job = execute(qiskit_qc, backend)
    result = job.result()
    test_unitary_2 = result.get_unitary(qiskit_qc).data
    end = time.time()

    print("Qiskit calculation time = %s" % (end - start))

    # get the unitary matrix from the result object
    print("Unitaries matching:", bool(norm(test_unitary_1 - test_unitary_2) < 1e-3))
    assert bool(norm(test_unitary_1 - test_unitary_2) < 1e-3)



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp.core.quantum_variable import QuantumVariable
from qrisp.core.session_merging_tools import merge
from qrisp.environments.quantum_environments import QuantumEnvironment
from qrisp.environments.quantum_inversion import invert
from qrisp.misc import lock, unlock


def temp_qv(function):
    def q_env_generator(*args, **kwargs):
        return ConjugationEnvironment(function, args, kwargs=kwargs)

    return q_env_generator


class CConjugationEnvironment(QuantumEnvironment):
    """
    This environment allows to temporarily compute a QuantumVariable which is available
    only inside this environment and automatically uncomputed once it is left.

    Similarly to the ConditionEnvironment, functions can be quickly transformed into
    this QuantumEnvironment using the temp_qv decorator.

    Parameters
    ----------
    qv_retrieval_function : function
        A function returning a QuantumVariable. It is not neccessary to uncompute
        intermediate quantum results of this function.
    args : list
        The list of arguments on which to call qv_retrieval_function.
    kwargs : dict, optional
        A dictionary of keyword arguments on which to call qv_retrieval_function.
        The default is {}.

    Examples
    --------

    We create a QuantumFloat, and temporary compute the tripled value. We then evaluate
    the condition of the tripled value being less than 6: ::

        from qrisp import temp_qv, QuantumFloat, QuantumBool, multi_measurement

        qf = QuantumFloat(5)
        q_bool = QuantumBool()

        qf[:] = {1 : 1, 2 : 1}

        @temp_qv
        def triple(qf):
            return 3*qf

        with triple(qf) as tripled_qf:

            with tripled_qf < 6:
                q_bool.flip()


    >>> print(multi_measurement([qf, q_bool]))
    {(1, True): 0.5, (2, False): 0.5}
    >>> print(tripled_qf)
    Exception: Tried to get measurement from deleted QuantumVariable


    """

    # Constructor of the class

    def __init__(self, qv_retrieval_function, args, kwargs={}):
        # Which
        self.qv_retrieval_function = qv_retrieval_function

        # Save the arguments on which the function should be evaluated
        self.args = args

        # Save the keyword arguments
        self.kwargs = kwargs

        # Note the QuantumSession of the arguments of the arguments
        self.arg_qs = merge(args)

        self.manual_allocation_management = True

    # Method to enter the environment
    def __enter__(self):
        self.user_qv = self.qv_retrieval_function(*self.args, **self.kwargs)

        super().__enter__()

        if not isinstance(self.user_qv, QuantumVariable):
            raise Exception("Retrieval function did not return a QuantumVariable")

        from qrisp import recursive_qv_search

        lock(recursive_qv_search(self.args))
        unlock(self.user_qv.reg)
        # perm_lock(self.user_qv.reg)

        merge(self.arg_qs, self.env_qs)

        return self.user_qv

    def __exit__(self, exception_type, exception_value, traceback):
        from qrisp import recursive_qv_search, redirect_qfunction

        unlock(recursive_qv_search(self.args))
        # perm_unlock(self.user_qv.reg)

        QuantumEnvironment.__exit__(self, exception_type, exception_value, traceback)

        redirected_function = redirect_qfunction(self.qv_retrieval_function)

        with invert():
            redirected_function(*self.args, target=self.user_qv, **self.kwargs)



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


import numpy as np


def auto_uncompute(*args, recompute=False):
    if len(args):
        return auto_uncompute_inner(args[0])
    else:
        if recompute:
            from qrisp import gate_wrap

            def auto_uncompute_helper(function):
                return gate_wrap(auto_uncompute_inner(function))

            return auto_uncompute_helper
        else:
            return auto_uncompute_inner(args[0])


# Decorator for auto uncomputed function
def auto_uncompute_inner(function):
    # Create auto uncomputed function
    def auto_uncomputed_function(*args, **kwargs):
        from qrisp.core import (
            QuantumVariable,
            multi_session_merge,
            recursive_qs_search,
            recursive_qv_search,
        )

        # Acquire quantum session

        qs_list = recursive_qs_search([args, kwargs])
        multi_session_merge(qs_list)

        if not len(qs_list):
            return function(*args, **kwargs)

        qs = qs_list[0]

        # Determine quantum variables to uncompute
        initial_qvs = set([hash(qv()) for qv in QuantumVariable.live_qvs])

        # Execute function
        result = function(*args, **kwargs)

        multi_session_merge(recursive_qs_search(result) + [qs])

        result_vars = set([hash(qv) for qv in recursive_qv_search(result)])

        uncomp_vars = []
        
        for qv in qs.qv_list:
            if not hash(qv) in initial_qvs.union(result_vars):
                uncomp_vars.append(qv)

        uncompute(qs, uncomp_vars)

        return result

    auto_uncomputed_function.__name__ = function.__name__ + "_auto_uncomputed"
    auto_uncomputed_function.__doc__ = function.__doc__
    # Return result
    return auto_uncomputed_function


verify = np.zeros(1)


def uncompute(qs, uncomp_vars, recompute=False):
    from qrisp import QuantumEnvironment

    if len(uncomp_vars) == 0:
        return
    
    temp_data = list(qs.data)
    qs.data = []
    
    
    
    for i in range(len(temp_data)):
        if isinstance(temp_data[i], QuantumEnvironment):
            env = temp_data[i]
            env.compile()
        else:
            qs.append(temp_data[i])
            
            
    uncomp_vars = list(set(uncomp_vars).intersection(qs.qv_list))
    qubits_to_uncompute = sum([qv.reg for qv in uncomp_vars], [])

    alloc_gates_remaining = list(qubits_to_uncompute)

    for i in range(len(qs.data)):
        instr = qs.data[i]
        if isinstance(instr, QuantumEnvironment):
            continue
        if instr.op.name == "qb_alloc" and instr.qubits[0] in alloc_gates_remaining:
            alloc_gates_remaining.remove(instr.qubits[0])

    if len(alloc_gates_remaining):
        incorrectly_allocated_qvs = []

        for qv in uncomp_vars:
            if set(alloc_gates_remaining).intersection(list(qv)):
                incorrectly_allocated_qvs.append(qv)

        incorrectly_allocated_qvs = [qv.name for qv in incorrectly_allocated_qvs]
        raise Exception(
            f"Could not uncompute QuantumVariables {incorrectly_allocated_qvs} "
            f"because they were not created within this QuantumEnvironment"
        )

    qc_to_uncompute = qs.copy()

    recompute_qubits = []

    for qb in qc_to_uncompute.qubits:
        if hasattr(qb, "recompute"):
            recompute_qubits.append(qb)

    from qrisp.uncomputation.unqomp import uncompute_qc

    uncomputed_qc = uncompute_qc(qc_to_uncompute, qubits_to_uncompute, recompute_qubits)

    from qrisp import QuantumCircuit

    qs.data = []
    QuantumCircuit.fast_append = True

    qs.data = uncomputed_qc.data

    for qv in uncomp_vars:
        # if recompute:
        # for qb in qv:
        # qb.recompute = True

        qv.delete(verify=bool(verify), recompute=recompute)

        # The uncomputation algorithm already added deallocation gates
        # Therefore we remove the deallocation gates
        for i in range(len(qv)):
            qv.qs.data.pop(-1)

    QuantumCircuit.fast_append = False



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp.quantum_network.qn_server import QuantumNetworkServer
from qrisp.quantum_network.quantum_network_session import QuantumNetworkSession
from qrisp import QuantumFloat, QuantumVariable, h, x
from qrisp import QuantumBool, QuantumChar

test = QuantumNetworkServer(socket_ip_address="127.0.0.1", background=True)
test.start()


charlie = QuantumNetworkSession(socket_ip="127.0.0.1", port=7070, name="charlie")
bob = QuantumNetworkSession(socket_ip="127.0.0.1", port=7070, name="bob")

print("Server creation done")
qf_1 = QuantumFloat(3, qs=bob)
qf_2 = QuantumFloat(3, qs=bob)

qf_1[:] = 2
qf_2[:] = 3

qf_3 = qf_2 * qf_1

qf_3 += 3

print("Compilation done")

bob.send_qv(qf_3, "charlie")


print("Communication done")

inbox = charlie.update()

print(inbox[0])


q_ch = QuantumBool()
q_ch[:] = False

bob.send_qv(q_ch, "charlie")

inbox = charlie.update()

print(inbox[0])



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp.interface import BackendClient


class VirtualBackend(BackendClient):
    """
    This class provides a virtual backend for circuit execution.
    Virtual means that the server is running on the same machine as a separate
    Python thread. This structure allows setting up convenient wrappers for foreign/
    modified circuit dispatching code.

    Circuits can be run using ``virtual_backend_instance.run(qc)``.
    The function that should be used to run a circuit can be specified during
    construction using the ``run_func`` parameter.


    Parameters
    ----------
    run_func : function
        A function that recieves a QuantumCircuit, an integer specifiying the amount of
        shots and a token in the form of a string. It returns the counts as a dictionary
        of bitstrings.
    name : str, optional
        A name for the virtual backend. The default is None.
    port : int, optional
        The port on which to listen. The default is None.
    ping_func : TYPE, optional
        A function which returns a BackendStatus object. The default is None.


    Examples
    --------


    We set up a VirtualBackend, which prints the received QuantumCircuit and returns the
    results of the QASM simulator. ::

        from qrisp.interface import convert_to_qiskit

        def run_func(qc, shots, token = ""):

            qiskit_qc = convert_to_qiskit(qc)

            print(qiskit_qc)

            from qiskit import Aer
            qiskit_backend = Aer.get_backend('qasm_simulator')

            #Run Circuit on the Qiskit backend
            return qiskit_backend.run(qiskit_qc, shots = shots).result().get_counts()


    >>> from qrisp.interface import VirtualBackend
    >>> example_backend = VirtualBackend(run_func)
    >>> from qrisp import QuantumFloat
    >>> qf = QuantumFloat(3)
    >>> qf[:] = 4
    >>> qf.get_measurement(backend = example_backend)
                 ┌─┐
    qf.0:   ─────┤M├──────
                 └╥┘┌─┐
    qf.1:   ──────╫─┤M├───
            ┌───┐ ║ └╥┘┌─┐
    qf.2:   ┤ X ├─╫──╫─┤M├
            └───┘ ║  ║ └╥┘
    cb_0: 1/══════╩══╬══╬═
                  0  ║  ║
    cb_1: 1/═════════╩══╬═
                     0  ║
    cb_2: 1/════════════╩═
    {4: 1.0}

    """

    def __init__(self, run_func, name=None, port=None, ping_func=None):
        if name is None:
            name = "virtual_quantum_backend"
        else:
            name = "virtual_quantum_backend_" + name

        from qrisp.interface import BackendServer

        # Create BackendServer
        self.backend_server = BackendServer(
            run_func, "::1", port=port, name=name, ping_func=ping_func
        )
        # Run the server (runs in the background)
        self.backend_server.start()
        # Connect client

        super().__init__(socket_ip="::1", port=port)

    def run(self, qc, shots, token=""):
        """
        Executes the function run_func specified at object creation.

        Parameters
        ----------
        qc : QuantumCircuit
            The QuantumCircuit to run.
        shots : int
            The amount of shots to perform.

        Returns
        -------
        res : dict
            A dictionary containing the measurement results.

        """
        return super().run(qc, shots, token)

    def ping(self):
        """
        Executes the function ping_func specified at object creation.


        Returns
        -------
        res : BackendStatus object

        """
        return super().ping()


class VirtualQiskitBackend(VirtualBackend):
    """
    This class instantiates a VirtualBackend using a Qiskit backend.
    This allows easy access to Qiskit backends through the qrisp interface.


    Parameters
    ----------
    backend : Qiskit backend object, optional
        A Qiskit backend object, which runs QuantumCircuits. The default is
        Aer.get_backend('qasm_simulator').
    port : int, optional
        The port to listen. The default is 8079.

    Examples
    --------

    We evaluate a QuantumFloat multiplication on the QASM-simulator.

    >>> from qrisp import QuantumFloat
    >>> from qrisp.interface import VirtualQiskitBackend
    >>> from qiskit import Aer
    >>> example_backend = VirtualQiskitBackend(backend = Aer.get_backend('qasm_simulator'))
    >>> qf = QuantumFloat(4)
    >>> qf[:] = 3
    >>> res = qf*qf
    >>> res.get_measurement(backend = example_backend)
    {9: 1.0}


    """

    from qiskit import Aer

    def __init__(self, backend=None, port=8079):
        if backend is None:
            from qiskit import Aer

            backend = Aer.get_backend("qasm_simulator")

        # Create the run method
        def run(qc, shots, token):
            # Convert to qiskit
            from qrisp.interface.circuit_converter import convert_circuit

            qiskit_qc = convert_circuit(qc, "qiskit", transpile=True)

            from qiskit import transpile

            qiskit_qc = transpile(qiskit_qc, backend=backend)
            # Run Circuit on the Qiskit backend
            qiskit_result = backend.run(qiskit_qc, shots=shots).result().get_counts()
            # Remove the spaces in the qiskit result keys
            result_dic = {}
            import re

            for key in qiskit_result.keys():
                counts_string = re.sub(r"\W", "", key)
                result_dic[counts_string] = qiskit_result[key]

            return result_dic

        # Create the ping method
        def ping():
            # Collect information about backend to create BackendStatus object

            config = backend.configuration()

            # Information about Qubits and their conznectivity
            from qrisp.interface import ConnectivityEdge, PortableQubit

            qubit_list = []
            for i in range(config.n_qubits):
                qubit_list.append(PortableQubit(str(i)))

            connectivity_map = []

            if config.coupling_map:
                for edge in config.coupling_map:
                    connectivity_map.append(
                        ConnectivityEdge(qubit_list[edge[0]], qubit_list[edge[1]])
                    )

            # Information about elementary gates
            from qrisp.circuit.standard_operations import op_list

            # TO-DO fix elementary ops
            for gate_name in config.basis_gates:
                for op in op_list:
                    if op().name == gate_name:
                        pass
                        # elementary_ops.append(op())

            elementary_ops = []
            from qrisp.interface import BackendStatus

            # Create BackendStatus object
            backend_status = BackendStatus(
                name=config.backend_name,
                qubit_list=qubit_list,
                elementary_ops=elementary_ops,
                connectivity_map=connectivity_map,
                online=True,
            )

            return backend_status

        # Call VirtualBackend constructor
        if isinstance(backend.name, str):
            name = backend.name
        else:
            name = backend.name()

        super().__init__(run, name=name, ping_func=ping, port=port)



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp.circuit.qubit import Qubit
from qrisp.core.library import mcx, p, rz, x
from qrisp.core.session_merging_tools import merge
from qrisp.environments import QuantumEnvironment
from qrisp.misc import perm_lock, perm_unlock, bin_rep


class ControlEnvironment(QuantumEnvironment):
    """
    This class behaves similarly to ConditionEnvironment but instead of a function
    calculating a truth value, we supply a list of qubits.
    The environment's content is then controlled on these qubits.

    An alias for this QuantumEnvironment is "control".


    Parameters
    ----------
    ctrl_qubits : list[Qubit]
        A list of qubits on which to control the environment's content.
    ctrl_state : int/str, optional
        The computational basis state which is supposed to activate the environment.
        Can be supplied as a bitstring or integer. The default is "1111..".


    Examples
    --------

    We create a QuantumVariable and control on some of it's qubits
    using the control alias ::

        from qrisp import QuantumVariable, QuantumString, multi_measurement, control, h

        qv = QuantumVariable(3)
        q_str = QuantumString()

        qv[:] = "011"
        h(qv[0])

        with control(qv[:2], "11"):
            q_str += "hello world"


    >>> print(multi_measurement([qv, q_str]))
    {('011', 'aaaaaaaaaaa'): 0.5, ('111', 'hello world'): 0.5}


    """

    def __init__(self, ctrl_qubits, ctrl_state=-1, ctrl_method=None, invert = False):
        self.arg_qs = merge(ctrl_qubits)

    
        self.ctrl_method = ctrl_method
        if isinstance(ctrl_qubits, Qubit):
            ctrl_qubits = [ctrl_qubits]
            
        if isinstance(ctrl_state, int):
            if ctrl_state < 0:
                ctrl_state += 2**len(ctrl_qubits)
                
            self.ctrl_state = bin_rep(ctrl_state, len(ctrl_qubits))[::-1]
        else:
            self.ctrl_state = str(ctrl_state)
        
            
        self.ctrl_qubits = ctrl_qubits
        self.invert = invert

        self.manual_allocation_management = True

        # For more information on why this attribute is neccessary check the comment
        # on the line containing subcondition_truth_values = []
        self.sub_condition_envs = []

        QuantumEnvironment.__init__(self)

    # Method to enter the environment
    def __enter__(self):
        from qrisp.qtypes.quantum_bool import QuantumBool

        if len(self.ctrl_qubits) == 1:
            self.condition_truth_value = self.ctrl_qubits[0]
        else:
            self.qbool = QuantumBool(name="ctrl_env*", qs=self.arg_qs)
            self.condition_truth_value = self.qbool[0]

        QuantumEnvironment.__enter__(self)

        merge(self.env_qs, self.arg_qs)

        return self.condition_truth_value

    def __exit__(self, exception_type, exception_value, traceback):
        from qrisp.environments import (
            ConditionEnvironment,
            ControlEnvironment,
            InversionEnvironment,
        )

        self.parent_cond_env = None

        QuantumEnvironment.__exit__(self, exception_type, exception_value, traceback)

        # Determine the parent environment
        for env in self.env_qs.env_stack[::-1]:
            if isinstance(env, (ControlEnvironment, ConditionEnvironment)):
                self.parent_cond_env = env
                break
            if not isinstance(env, (QuantumEnvironment, InversionEnvironment)):
                break

    def compile(self):
        from qrisp import QuantumBool
        from qrisp.environments import ConditionEnvironment

        # Create the quantum variable where the condition truth value should be saved
        # Incase we have a parent environment we create two qubits because
        # we use the second qubit to compute the toffoli of this one and the parent
        # environments truth value in order to not have the environment operations
        # controlled on two qubits

        if len(self.env_data):
            # The first step we have to perform is calculating the truth value of the
            # environments quantum condition. For this we differentiate between
            # the case that this condition is embedded in another condition or not

            ctrl_qubits = list(self.ctrl_qubits)
            cond_compile_ctrl_state = self.ctrl_state
            

            if self.parent_cond_env is not None:
                
                # In the parent case we also need to make sure that the code is executed
                # if the parent environment is executed. A possible approach would be
                # to control the content on both, the parent and the chield truth value.
                # However, for each nesting level the gate count to generate
                # the controlled-controlled-controlled... version of the gates inside
                # the environment increases exponentially. Because of this we compute
                # the toffoli of the parent and child truth value
                # and controll the environment gates on this qubit.

                # Synthesize the condition of the environment
                # into the condition truth value qubit
                if len(ctrl_qubits) == 1:
                    from qrisp.misc import retarget_instructions

                    self.qbool = QuantumBool(name="ctrl_env*")
                    retarget_instructions(
                        self.env_data, [self.condition_truth_value], [self.qbool[0]]
                    )

                    if len(self.env_qs.data):
                        if isinstance(self.env_qs.data[-1], QuantumEnvironment):
                            env = self.env_qs.data.pop(-1)
                            env.compile()

                    self.condition_truth_value = self.qbool[0]

                ctrl_qubits.append(self.parent_cond_env.condition_truth_value)
                
                
                parent_ctrl_state = "1"
                
                if isinstance(self.parent_cond_env, ControlEnvironment):
                    if len(self.parent_cond_env.ctrl_qubits) == 1:
                        if self.parent_cond_env.ctrl_state == "0":
                            parent_ctrl_state = "0"
                
                    if self.parent_cond_env.invert:
                        if parent_ctrl_state == "0":
                            parent_ctrl_state  = "1"
                        elif parent_ctrl_state == "1":
                            parent_ctrl_state  = "0"
                        
                cond_compile_ctrl_state = cond_compile_ctrl_state + parent_ctrl_state
                    
                

            if len(ctrl_qubits) > 1:
                
                if len(ctrl_qubits) > 5:
                    method = "auto"
                else:
                    method = "gray_pt"
                
                mcx(
                    ctrl_qubits,
                    self.condition_truth_value,
                    ctrl_state=cond_compile_ctrl_state,
                    method=method,
                )

            perm_lock(ctrl_qubits)
            # unlock(self.condition_truth_value)

            # This list will contain the qubits holding the truth values of
            # conditional/control environments within this environment.
            # The instruction from the subcondition environments do not need to be
            # controlled, since their compile method compiles their condition
            # truth value based on the truth value of the parent environment.
            subcondition_truth_values = [
                env.condition_truth_value for env in self.sub_condition_envs
            ]
            inversion_tracker = 1
            

            
            # Now we need to recover the instructions from the data list
            # and perform their controlled version on the condition_truth_value qubit
            while self.env_data:
                instruction = self.env_data.pop(0)

                # If the instruction == conditional environment, compile the environment
                if isinstance(instruction, (ControlEnvironment, ConditionEnvironment)):
                    instruction.compile()

                    subcondition_truth_values = [
                        env.condition_truth_value for env in self.sub_condition_envs
                    ]
                    continue

                # If the instruction is a general environment, compile the instruction
                # and add the compilation result to the list of instructions
                # that need to be conditionally executed.
                elif issubclass(instruction.__class__, QuantumEnvironment):
                    temp_data_list = list(self.env_qs.data)
                    self.env_qs.clear_data()
                    instruction.compile()
                    self.env_data = list(self.env_qs.data) + self.env_data
                    self.env_qs.clear_data()
                    self.env_qs.data.extend(temp_data_list)

                    subcondition_truth_values = [
                        env.condition_truth_value for env in self.sub_condition_envs
                    ]
                    continue

                if (
                    instruction.op.name in ["qb_alloc", "qb_dealloc"]
                    and instruction.qubits[0] != self.condition_truth_value
                ) or instruction.op.name == "barrier":
                    self.env_qs.append(instruction)
                    continue

                if set(instruction.qubits).intersection(subcondition_truth_values):
                    self.env_qs.append(instruction)
                    continue

                # Support for inversion of the condition without opening a new
                # environment
                # if set(instruction.qubits).issubset(self.user_exposed_qbool):
                if set(instruction.qubits).issubset([self.condition_truth_value]):
                    if instruction.op.name == "x":
                        inversion_tracker *= -1
                        x(self.condition_truth_value)
                    elif instruction.op.name == "p":
                        p(instruction.op.params[0], self.condition_truth_value)
                    elif instruction.op.name == "rz":
                        rz(instruction.op.params[0], self.condition_truth_value)
                    else:
                        raise Exception(
                            "Tried to perform invalid operations"
                            "on condition truth value (allowed are x, p, rz)"
                        )
                    continue
                
                if len(ctrl_qubits) == 1:
                    ctrl_state = self.ctrl_state
                else:
                    ctrl_state = "1"
                    
                if self.invert:
                    
                    new_ctrl_state = ""
                    for c in ctrl_state:
                        if c == "1":
                            new_ctrl_state += "0"
                        else:
                            new_ctrl_state += "1"
                            
                    ctrl_state = new_ctrl_state
                    
                # Create controlled instruction
                instruction.op = instruction.op.control(
                    num_ctrl_qubits=1, method=self.ctrl_method, ctrl_state=ctrl_state
                )

                # Add condition truth value qubit to the instruction qubit list
                instruction.qubits = [self.condition_truth_value] + list(
                    instruction.qubits
                )
                # Append instruction
                self.env_qs.append(instruction)

            perm_unlock(ctrl_qubits)
            
            if inversion_tracker == -1:
                x(self.condition_truth_value)

            if len(ctrl_qubits) > 1:
                
                if len(ctrl_qubits) > 5:
                    method = "auto"
                else:
                    method = "gray_pt_inv"
                
                mcx(
                    ctrl_qubits,
                    self.qbool,
                    method=method,
                    ctrl_state=cond_compile_ctrl_state,
                )
                self.qbool.delete()

        if self.parent_cond_env is not None:
            self.parent_cond_env.sub_condition_envs.extend(
                self.sub_condition_envs + [self]
            )


control = ControlEnvironment



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


import numpy as np
import sympy as sp

from qrisp.misc import int_as_array
from qrisp.circuit import Operation
#from qrisp import int_encoder, get_ordered_symbol_list


# from qiskit import *


# Class to describe truth tables
# Can be intialized with a list of bitstrings, or a numpy array with 1 and 0s or a
# sympy expression
class TruthTable:
    def __init__(self, init_object):
        # This handles the case that the initializing object is a sympy expression
        # the variables will be sorted alphabetically regarding their name
        if isinstance(init_object, sp.Expr):
            expr = init_object

            init_str = ""

            self.expr = expr
            from qrisp.arithmetic.poly_tools import get_ordered_symbol_list

            # Retrieve a list of symbols in a given expression
            # Reverse to get the correct endian convention for qiskit
            symbol_list = get_ordered_symbol_list(expr)[::-1]

            # Replace the variables in the expression with 1s and 0s and add them to the
            # intialization string
            from sympy import lambdify

            expr_func = lambdify([symbol_list], expr, "numpy")

            for i in range(2 ** len(symbol_list)):
                if len(symbol_list):
                    symbol_const = int_as_array(i, len(symbol_list))
                else:
                    symbol_const = []

                res = expr_func(symbol_const) % 2
                if not set({int(res)}) <= set([0, 1]):
                    raise Exception("Sympy expression returned non-boolean value")

                init_str += str(int(res))
            # this new object will now contain the string list which will be further
            # prosecuted in this function
            init_object = [init_str]

        # This handles the case that the truth table is instantiated as a list of
        # strings
        if isinstance(init_object, list):
            # Check if the list contains only bit strings of the same length
            for i in range(1, len(init_object)):
                if len(init_object[0]) != len(init_object[i]):
                    raise Exception(
                        "Tried to initialize truth table with varying lengths"
                    )

            # Save string list into object structure
            self.str_l = init_object

            # Generate numpy array representation
            self.n_rep = [[int(c) for c in singular_tt] for singular_tt in init_object]
            self.n_rep = np.array(self.n_rep).transpose()

        # This handles the case that the truth table is instantiated as a numpy array
        if isinstance(init_object, np.ndarray):
            # Save array into object structure
            self.n_rep = init_object

            # Generate list of bitstrings
            self.str_l = [
                "".join(
                    [str(int(init_object[i, j])) for i in range(init_object.shape[0])]
                )
                for j in range(init_object.shape[1])
            ]

        # Check some conditions
        if not (set(self.n_rep.flatten()) <= set([0, 1])):
            raise Exception("Tried to initialize truth table with a non boolean values")

        # Set shortcuts to some prevalent parameters of the truth table
        self.shape = self.n_rep.shape

        if not (int(np.log2(self.shape[0])) == np.log2(self.shape[0])):
            raise Exception(
                "Tried to initialize truth table with a boolean function with a "
                "length which is not an integer power of 2"
            )

        self.bit_amount = int(np.log2(self.shape[0]))

    # Returns the "left side" of the truth table, ie. the variable values
    # (starting with 00..00, 00..01, 00..10,...)
    def variable_array(self):
        variable_table_array = [
            int_as_array(i, int(np.log2(self.shape[0]))) for i in range(self.shape[0])
        ]
        variable_table_array = np.array(variable_table_array)

        return TruthTable(variable_table_array)

    # Returns the cofactors of the truth table
    # If the truth table describes a boolean function f(x0,x1,x2...)
    # Then the cofactors for index 2 is f(x0,x1,0,...) and f(x0,x1,1,...)
    def cofactors(self, i):
        # Set up lists for the values of the cofactor tables
        cofactor_tables = [[], []]

        # Iterate over all value constellations and put them either
        # into the "first" or the "second" cofactor
        for k in range(self.shape[0]):
            if int_as_array(k, self.bit_amount)[i]:
                cofactor_tables[1].append(self.n_rep[k, :])
            else:
                cofactor_tables[0].append(self.n_rep[k, :])

        return [
            TruthTable(np.array(cofactor_tables[0])),
            TruthTable(np.array(cofactor_tables[1])),
        ]

    # Returns a single column as a truth tables
    def sub_table(self, i):
        return TruthTable(np.array([self.n_rep[:, i]]).transpose())

    # Method to print truth tables
    def __str__(self):
        return str(np.array(self.n_rep))

    # Swaps to columns of the truth table
    def swap_col(self, i, j):
        temp_array = np.array(self.n_rep)
        temp_col = self.n_rep[:, i]
        temp_array[:, i] = temp_array[:, j]
        temp_array[:, j] = temp_col
        return TruthTable(temp_array)

    # Calculates the complexity of the truth table (more to that in the definition
    # of "D")
    def calc_complexity(self, as_array=False):
        if as_array:
            return np.array([D(self.n_rep[:, i]) for i in range(len(self.str_l))])

        c = 0

        for i in range(len(self.str_l)):
            c += D(self.n_rep[:, i])

        return -c

    # Synthesizes a ciruict which represents the truth table
    def q_synth(self, input_var, output_var, method="gray"):
        if output_var.size != self.shape[1]:
            raise Exception(
                "Given output variable doesn't include the required amount of qubits"
            )

        if len(input_var) != self.bit_amount:
            raise Exception(
                "Given input variable doesn't include the required amount of qubits"
            )

        # Use gray synthesis to synthesize truth table
        if method == "gray":
            from qrisp.logic_synthesis.gray_synthesis import gray_logic_synth

            gray_logic_synth(input_var, output_var, self, phase_tolerant=False)

        # Use phase tolerant gray synthesis to synthesize truth table
        elif method == "gray_pt":
            from qrisp.logic_synthesis.gray_synthesis import gray_logic_synth

            gray_logic_synth(input_var, output_var, self, phase_tolerant=True)

        elif method == "gray_pt_inv":
            from qrisp.logic_synthesis.gray_synthesis import gray_logic_synth
            from qrisp.misc import quantum_invert

            quantum_invert(
                gray_logic_synth, [input_var, output_var, self, True], input_var.qs
            )

        elif method == "pprm_pt":
            from qrisp.logic_synthesis.pprm_synthesis import pprm

            pprm(input_var, output_var, self, phase_tolerant=True)

        elif method == "pprm":
            from qrisp.logic_synthesis.pprm_synthesis import pprm

            pprm(input_var, output_var, self, phase_tolerant=False)

        elif method == "best":
            input_var.qs.append(
                self.gate_synth(), list(input_var.reg) + list(output_var.reg)
            )

        else:
            raise Exception("Given synthesis method unknown")

        # Append the truth table to the generated operation, so that it can be
        # resynthesized phase tolerantly during uncomputation
        # input_var.qs.data[-1].op.tt = self
        # input_var.qs.data[-1].op.logic_synth_method = method

        input_var.qs.data[-1].op = LogicSynthGate(
            input_var.qs.data[-1].op, self, phase_tolerant=method
        )

    def gate_synth(self, method="best", inv=False):
        from qrisp.core import QuantumSession, QuantumVariable

        if method == "best":
            qs_list = []
            methods = ["gray_pt", "td", "td_pk", "pprm_pt"]

            if self.bit_amount != 1:
                methods = ["gray_pt", "pprm_pt"]
            else:
                methods = ["pprm_pt"]

            for m in methods:
                qs = QuantumSession()
                input_var = QuantumVariable(self.bit_amount, qs)
                output_var = QuantumVariable(self.shape[1], qs)

                self.q_synth(input_var, output_var, method=m)

                qs_list.append(qs)

            cnot_list = [qs_list[i].cnot_count() for i in range(len(methods))]

            qs = qs_list[cnot_list.index(min(cnot_list))]

        else:
            qs = QuantumSession()

            input_var = QuantumVariable(self.bit_amount, qs)
            output_var = QuantumVariable(self.shape[1], qs)

            self.q_synth(input_var, output_var, method=method)

        if not inv:
            return qs.to_gate()
        else:
            return qs.to_gate().inverse()

    def __or__(self, other):
        if self.shape != other.shape:
            raise Exception("Tried to or two truth tables of different shape")

        new_tt = np.zeros(self.shape)
        for i in range(self.shape[0]):
            for j in range(self.shape[1]):
                new_tt[i, j] = self.n_rep[i, j] or other.n_rep[i, j]

        return TruthTable(new_tt)


T_bib = {}


# Implementation of the truth table complexity measure described in Miller, D. M.,
# “Spectral and Two-Place Decomposition Techniques in Reversible Logic,”
# Proc. Midwest Symposium on Circuits and Systems, on CD-ROM, August 2
def T(n):
    global T_bib

    if n in T_bib.keys():
        return T_bib[n]

    if n == 0:
        return np.array([[1]])

    result = np.zeros((2**n, 2**n))

    temp = T(n - 1)

    result[: 2 ** (n - 1), : 2 ** (n - 1)] = temp
    result[2 ** (n - 1) :, : 2 ** (n - 1)] = temp
    result[: 2 ** (n - 1), 2 ** (n - 1) :] = temp
    result[2 ** (n - 1) :, 2 ** (n - 1) :] = -temp

    T_bib[n] = result
    return result


def fwht(a):
    """In-place Fast Walsh–Hadamard Transform of array a."""
    h = 1
    while h < len(a):
        for i in range(0, len(a), h * 2):
            for j in range(i, i + h):
                x = a[j]
                y = a[j + h]
                a[j] = x + y
                a[j + h] = x - y
        h *= 2


def rw_spectrum(f):
    if isinstance(f, str):
        f = [int(c) for c in f]
        f = np.array(f)

    a = f.copy()
    fwht(a)
    return a

    size = len(f)

    if np.log2(size) != int(np.log2(size)):
        raise Exception(
            "The given function does not have the length to properly represent "
            "a truth table"
        )

    n = int(np.log2(size))

    return np.dot(T(n), f)


def C(f):
    size = len(f)

    if np.log2(size) != int(np.log2(size)):
        raise Exception(
            "The given function does not have the length to properly represent "
            "a truth table"
        )

    n = int(np.log2(size))

    rw_spec = rw_spectrum(f)
    sum_ = 0
    for i in range(size):
        sum_ += sum(int_as_array(i, n)) * rw_spec[i] ** 2

    sum_ = sum_ / 2 ** (n - 2)

    return 1 / 2 * (n * size - sum_)


def NZ(f):
    R = rw_spectrum(f)

    sum_ = 0
    for i in R:
        if i == 0:
            sum_ += 1
    return sum_


def D(f):
    size = len(f)
    if np.log2(size) != int(np.log2(size)):
        raise Exception(
            "The given function does not have the length to properly represent "
            "a truth table"
        )

    n = int(np.log2(size))

    return int(n * 2 ** (n - 3) * NZ(f) + C(f))


def synth_poly(truth_table, column=0, coeff=None):
    if coeff is None:
        coeff = sp.symbols(
            "".join([" x" + str(i) for i in range(truth_table.bit_amount)])
        )
        if truth_table.bit_amount == 1:
            coeff = [coeff]

    try:
        expr = truth_table.expr
        symbols = get_ordered_symbol_list(expr)
        coeff_temp = sp.symbols(
            "".join(
                [" abcdefghijkllmn" + str(i) for i in range(truth_table.bit_amount)]
            )
        )
        subs_dic = {symbols[i]: coeff_temp[i] for i in range(len(coeff))}
        expr = expr.subs(subs_dic)
        subs_dic = {coeff_temp[i]: coeff[i] for i in range(len(coeff))}
        expr = expr.subs(subs_dic).expand()
        # print(subs_dic)
        # print(sp.Poly(expr.subs(subs_dic).expand(), domain = sp.GF(2)))
        return sp.Poly(expr.subs(subs_dic).expand(), domain=sp.GF(2)).expr
    except:
        pass

    poly = sp.sympify(0)
    for i in range(truth_table.shape[0]):
        if truth_table.n_rep[i, column]:
            temp = sp.sympify(1)
            array = int_as_array(i, truth_table.bit_amount)
            for j in range(len(array)):
                if array[j]:
                    temp *= coeff[-j - 1]
                else:
                    temp *= coeff[-j - 1] + 1
            poly += temp

    if poly == 0:
        return poly
    return sp.Poly(poly.expand(), domain=sp.GF(2)).expr
    # return filter_pow(sp.Poly(poly.expand(), domain = sp.GF(2)).expr)


class LogicSynthGate(Operation):
    def __init__(self, init_op, tt, phase_tolerant=False):
        self.tt = tt

        self.logic_synth_method = phase_tolerant

        for var in init_op.__dict__.keys():
            if var != "inverse":
                self.__dict__[var] = init_op.__dict__[var]

    def inverse(self):
        return LogicSynthGate(Operation.inverse(self), self.tt, self.logic_synth_method)


def check_synthesis(tt, gate, log_output=False):
    synth_correct = True
    from qrisp.core import QuantumSession, QuantumVariable

    for i in range(tt.shape[0]):
        qs = QuantumSession()
        input_var = QuantumVariable(tt.bit_amount, qs)
        output_var = QuantumVariable(gate.num_qubits - tt.bit_amount, qs)
        int_encoder(input_var, i)
        qs.append(gate, range(gate.num_qubits))

        tt_value = "".join([str(x) for x in tt.n_rep[i]])
        synth_res = list(output_var.get_measurement().keys())[0]

        if log_output:
            print("TruthTable:", tt_value)
            print("Synthesis:", synth_res)
            print("---")

        synth_correct = synth_correct and tt_value == synth_res

    return synth_correct



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


import threading

from thrift.protocol import TBinaryProtocol
from thrift.transport import TSocket, TTransport

from qrisp.interface.thrift_interface.stoppable_thrift_server import (
    StoppableThriftServer,
)
from qrisp.interface.thrift_interface.codegen import BackendService


# Returns the hosts ip
def get_ip():
    import socket

    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    try:
        # doesn't even have to be reachable
        s.connect(("10.255.255.255", 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = "127.0.0.1"
    finally:
        s.close()
    return IP


# This class describes a Backend Server
class BackendServer:
    """
    This class allows convenient setup of a server respecting the Qrisp interface.

    Parameters
    ----------
    run_func : function
        A function that receives a QuantumCircuit, an integer specifying the amount of
        shots and a token in the form of a string. It returns the counts as a dictionary
        of bitstrings.
    socket_ip_address : str, optional
        The IP address of where the listening socket should be opened. By default, the
        IP address of the hosting machine will be used.
    port : int, optional
        The port on which to listen for requests. By default, 9090 will be used.
    name : str, optional
        A name for the server. The default is "generic_quantum_backend_server".
    ping_func : function, optional
        A function returning a BackendStatus object. The default is None.
    online_status : bool, optional
        A bool specifying whether the server should be able to handle requests directly
        after startup. The default is True.


    Examples
    --------

    We create a server listening on the localhost IP address using a run function which
    prints the token and queries the QASM-simulator. ::



        def run_func(qc, shots, token):

            #Convert to qiskit
            from qrisp.interface.circuit_converter import convert_circuit
            qiskit_qc = convert_circuit(qc, "qiskit")

            print(token)

            from qiskit import Aer
            qiskit_backend = Aer.get_backend('qasm_simulator')

            #Run Circuit on the Qiskit backend
            return qiskit_backend.run(qiskit_qc, shots = shots).result().get_counts()

        from qrisp.interface import BackendServer
        example_server = BackendServer(run_func, socket_ip_address = "127.0.0.1", port = 8080)
        example_server.start()


    """

    def __init__(
        self,
        run_func,
        socket_ip_address=None,
        port=None,
        transport=None,
        name="generic_quantum_backend_server",
        ping_func=None,
        online_status=True,
    ):
        if socket_ip_address is None:
            socket_ip_address = get_ip()

        self.transport = transport

        self.socket_ip_address = socket_ip_address

        self.online_status = online_status

        self.run_func = run_func

        if ping_func is None:

            def ping_func():
                from qrisp.interface import BackendStatus

                return BackendStatus(name=name)

        self.ping_func = ping_func

        self.name = name

        self.port = port

    # Starts the server
    def start(self):
        """
        Starts the server.
        """

        if self.port is None:
            self.port = 9010

        self.thread_name = self.name + ":" + str(self.port)

        # Create the BackendServiceHandler class
        ping_func = self.ping_func
        run_func = self.run_func
        pass_online_status_bool_by_reference = lambda: self.online_status

        class BackendServiceHandler:
            def run(self, qc, shots, token):
                return run_func(qc, shots, token)

            def ping(self):
                status = ping_func()
                status.online = pass_online_status_bool_by_reference()
                return status

        # Create server
        handler = BackendServiceHandler()
        processor = BackendService.Processor(handler)
        tfactory = TTransport.TBufferedTransportFactory()

        if self.transport is None:
            self.transport = TSocket.TServerSocket(
                host=self.socket_ip_address, port=self.port
            )

        pfactory = TBinaryProtocol.TBinaryProtocolFactory()

        # server = TServer.TSimpleServer(processor, transport, tfactory, pfactory)
        # self.server = TServer.TThreadedServer(processor, self.transport, tfactory, pfactory, daemon = True)

        # self.server = TNonblockingServer.TNonblockingServer(processor, self.transport, tfactory, pfactory)
        self.server = StoppableThriftServer(
            processor, self.transport, tfactory, pfactory, daemon=True
        )

        # Create thread
        self.thr = threading.Thread(target=self.server.serve, name=self.thread_name)
        self.thr.setDaemon(True)

        # Start the thread
        self.thr.start()

        while True:
            if self.server.is_running:
                break

    def stop(self):
        """
        Stops the server.
        """
        if hasattr(self, "server"):
            self.server.stop()
            self.thr.join()

    def __del__(self):
        self.stop()



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


import numpy as np
import sympy as sp

from qrisp.arithmetic.poly_tools import (
    expr_to_list,
    filter_pow,
    get_ordered_symbol_list,
)
from qrisp.core import QuantumArray, QuantumVariable, cp, cx, cz, h, mcx, p
from qrisp.misc import gate_wrap

# Threshold of rounding used in detecting integer multiples of pi
pi_mult_round_threshold = 11

depth_tracker = {}


# Check if the given expression is a polynomial
def check_for_polynomial(expr):
    for x in sp.preorder_traversal(expr):
        if not isinstance(
            x,
            (
                sp.core.mul.Mul,
                sp.core.add.Add,
                sp.core.symbol.Symbol,
                sp.core.numbers.Integer,
                sp.core.numbers.Float,
                sp.core.power.Pow,
            ),
        ):
            return False

        if isinstance(x, sp.core.power.Pow):
            if not isinstance(x.args[1], sp.core.numbers.Integer):
                return False
    return True


# Efficient implementation of the multicontrolled U_g gate
def multi_controlled_U_g(
    output_qf, control_qb_list, y, phase_tolerant=False, use_gms=False
):
    # Set alias for quantum session
    qs = output_qf.qs

    # For one control qubit, there is no advantage in calculating an ancilla qubit
    if len(control_qb_list) == 1:
        ancilla = control_qb_list

    # Otherwise request an ancilla qubit
    else:
        ancilla = QuantumVariable(1, qs=output_qf.qs, name="sbp_anc*")

    # If an ancilla qb is available
    # we synthesize the result of the boolean multiplication in it

    if len(control_qb_list) != 1:
        if use_gms:
            toffoli_method = "gms"
        else:
            toffoli_method = "gray_pt"

        # Apply multi-controlled x gate
        mcx(control_qb_list, ancilla[0], method=toffoli_method)

    # Now we execute the phase gates
    # For this there is 2 possibilies
    # Either use regular phase gates or use GMS gates
    # GMS gates are ion-trap native gates, that allow the entanglement
    # of multiple qubits within a single laser pulse
    if use_gms:
        # GMSEnvironment is an environment that allows programming
        # with regular phase gates but converts everything programmed
        # into a GMS gate upon leaving the environment
        from qrisp.environments.GMSEnvironmentironment import GMSEnvironment

        env = GMSEnvironment(qs)

    else:
        # This is an environment that simply does nothing
        from qrisp.environments.quantum_environments import QuantumEnvironment

        env = QuantumEnvironment()

    # Enter environment
    # if True:
    env.manual_allocation_management = True
    with env:
        phase_accumulator = 0
        cz_counter = 0

        # Execute controlled rotations
        for i in range(output_qf.size):
            # The -i (instead of i) reverses the gate order implying that
            # we can leave out the swaps of the QFT

            rot_angle = 2 * np.pi * 2 ** (-i - 1) * y
            rot_angle = rot_angle % (2 * np.pi)

            if (
                np.round(rot_angle, pi_mult_round_threshold) != 0
                and np.round((-rot_angle) % (2 * np.pi), pi_mult_round_threshold) != 0
            ):
                # If phase is equal to pi, execute cz gate (costs one CNOT less than CP)
                if False:
                    # if np.round(abs(rot_angle) - np.pi, pi_mult_round_threshold) == 0:
                    cz(ancilla[0], output_qf.reg[i])
                    cz_counter += 1

                    phase_accumulator += rot_angle / 2
                # Otherwise execute cp gate
                else:
                    cx(ancilla[0], output_qf.reg[i])
                    p(-rot_angle / 2, output_qf.reg[i])
                    cx(ancilla[0], output_qf.reg[i])
                    p(rot_angle / 2, output_qf.reg[i])

                    phase_accumulator += rot_angle / 2
                    # cp(rot_angle, ancilla_qb, output_qf.reg[i])
                    # crz(rot_angle, ancilla[0], output_qf.reg[i])
        p(phase_accumulator - np.pi * cz_counter / 2, ancilla[0])

    # Uncompute boolean multiplication
    if len(control_qb_list) != 1:
        if not use_gms:
            toffoli_method += "_inv"

        mcx(control_qb_list, ancilla[0], method=toffoli_method)
        # gray_logic_synth_qb_list(control_qb_list,
        # ancilla_qb, ancilla_qv.qs, tt, inverse = True)
        ancilla.delete()

    return phase_accumulator


# This function returns the U_g gate (up to qiskit endian conventions)
# This is used for applying U_g gates which have no control-knobs
def U_g(y, qv):
    for i in range(len(qv)):
        rot_angle = rot_angle = np.pi * 2 ** (-i) * y
        rot_angle = rot_angle % (2 * np.pi)
        if (
            np.round(rot_angle, pi_mult_round_threshold) != 0
            and np.round((-rot_angle) % (2 * np.pi), pi_mult_round_threshold) != 0
        ):
            p(rot_angle, qv[i])


# This function encodes a semi-boolean polynomial into a circuit i.e.
# For the polynomial p(x_0, x_1, x_2) = 4*x_0*x_1 + 2*x_0*x_2
# the effect of this function is:
# U_f |x_0, x_1, x_2>|0> = |x_0, x_1, x_2>|p(x_1,x_2,x_3)>
def sb_polynomial_encoder(
    input_qf_list, output_qf, poly, inplace_mult=1, use_gms=False, init_op="auto"
):
    # As the polynomial has only boolean variables,
    # powers can be ignored since x**k = x for x in GF(2)
    poly = filter_pow(poly.expand()).expand() / 2.0**output_qf.exponent

    # Acquire list of symbols present in the polynomial
    symbol_list = []

    for qf in input_qf_list:
        temp_var_list = list(sp.symbols(qf.name + "_" + "0:" + str(qf.size)))
        symbol_list += temp_var_list

    n = len(symbol_list)

    if n != sum([var.size for var in input_qf_list]):
        raise Exception(
            "Input variables do not the required amount of qubits to encode polynomial"
        )

    # Acquire monomials in list form
    monomial_list = expr_to_list(poly)

    from qrisp import QFT, conjugate, QuantumEnvironment
    

    if inplace_mult == 1:
        env = conjugate(QFT)(
            output_qf,
            inv=False,
            exec_swap=False,
            inplace_mult=inplace_mult,
            use_gms=use_gms,
        )
    else:
        QFT(
            output_qf,
            inv=False,
            exec_swap=False,
            inplace_mult=inplace_mult,
            use_gms=use_gms,
        )
        env = QuantumEnvironment()

    with env:
        # The list of qubits contained in the variables of input_var_list
        input_qubits = sum([list(var.reg) for var in input_qf_list], [])
    
        control_qubit_list = []
        y_list = []
    
        # Iterate through the monomials
        for monom in monomial_list:
            # Prepare the two variables coeff (which is the coefficient of the monomial)
            # And the list of variables which appear in the monomial
    
            # For this, go through the cases which can appear
    
            # This describes the case where there is only a single term in the monomial
            # Either a constant or a variable
            if len(monom) == 1:
                if isinstance(monom[0], sp.core.symbol.Symbol):
                    coeff = 1
                    variables = list(monom)
                else:
                    coeff = float(monom[0])
                    variables = []
    
            # This describes the case where there is multiple terms in the monomial
            elif not isinstance(monom[0], sp.core.symbol.Symbol):
                coeff = monom[0]
                variables = list(monom[1:])
            else:
                coeff = 1
                variables = list(monom)
    
            # Check if the coefficient is an integer (up to float errors)
            if abs(int(np.round(float(coeff))) - coeff) > 1e-14:
                pass
                # raise Exception("Tried to encode sb-polynomial
                # with non-integer coefficient")
    
            # Append coefficient to y_list
            y_list.append(int(np.round(float(coeff))))
    
            # Prepare the qubits on which the U_g should be controlled
            control_qubit_numbers = [symbol_list.index(var) for var in variables]
    
            control_qubits = [input_qubits[nr] for nr in control_qubit_numbers]
    
            control_qubits = list(set(control_qubits))
    
            control_qubits.sort(key=lambda x: x.identifier)
    
            control_qubit_list.append(control_qubits)
    
        # Now we apply the multi controlled U_g gate
        # Here the order in which they are applied makes a huge difference
        # In order to determine, which U_g to apply next, we evaluate a cost function
        # (which we determined through trial and error)
        # and choose the U_g with the lowest cost
    
        # Note that for this feature to yield an improvement, the quantum session requires
        # multiple free ancilla qubits to work on
        def delay_cost(depth_array):
            return max(depth_array)
    
        def find_best_monomial(control_qubit_list, depth_dic):
            delay_cost_list = []
    
            for i in range(len(control_qubit_list)):
                depth_array = np.array([depth_dic[qb] for qb in control_qubit_list[i]])
    
                delay_cost_list.append(delay_cost(depth_array))
    
            return np.argmin(delay_cost_list)
    
        # Iterate through the list of U_g gates
        while control_qubit_list:
            # TO-DO fix depth calculation inside environment
            # Update depth_dic (contains the depth of each qubit)
            # depth_dic = output_qf.qs.get_depth_dic()
    
            # Determine best U_g
            # monomial_index = find_best_monomial(control_qubit_list, depth_dic)
            monomial_index = 0
            # Find control qubits and their coefficient
            control_qubits = control_qubit_list.pop(monomial_index)
            y = y_list.pop(monomial_index)
    
            # Apply (controlled) U_g
            if len(control_qubits):
                multi_controlled_U_g(output_qf, control_qubits, y, use_gms=use_gms)
            else:
                U_g(y, output_qf)

    if inplace_mult != 1:
        # Apply QFT
        QFT(output_qf, inv=True, exec_swap=False, use_gms=use_gms)


# Multiplies two integers
# The general idea is to represent the integers as values of two polynomials
# p(x) = x_0 + 2*x_1 + 4*x_2 ...
# p(y) = y_0 + 2*y_1 + 4*y_2 ...
# these polynomials get multiplied and which forms a bigger polynomial
# which can be encoded using the polynomial encoder
def sbp_mult(factor_1_qf, factor_2_qf, output_qf=None):
    """
    Performs multiplication based on the evaluation of
    `semi-boolean polynomials <https://ieeexplore.ieee.org/document/9815035>`_.

    Parameters
    ----------
    factor_1_qf : QuantumFloat
        The first factor to multiply.
    factor_2_qf : QuantumFloat
        The second factor to multiply.
    output_qf : QuantumFloat, optional
        The QuantumFloat to store the result in.
        By default, a suited new QuantumFloat is created.

    Returns
    -------
    output_qf : QuantumFloat
        A QuantumFloat containing the result of the multiplication.

    Examples
    --------

    We multiply two QuantumFloats:

    >>> from qrisp import QuantumFloat, sbp_mult
    >>> qf_0 = QuantumFloat(3)
    >>> qf_1 = QuantumFloat(3)
    >>> qf_0[:] = 3
    >>> qf_1[:] = 4
    >>> qf_res = sbp_mult(qf_0, qf_1)
    >>> print(qf_res)
    {12: 1.0}
    """

    if output_qf is None:
        from qrisp.qtypes.quantum_float import create_output_qf

        output_qf = create_output_qf([factor_1_qf, factor_2_qf], op="mul")
    # Multiply the polynmials
    mult_poly = factor_1_qf.sb_poly(output_qf.msize) * factor_2_qf.sb_poly(
        output_qf.msize
    )

    # Apply sb encoder
    sb_polynomial_encoder([factor_1_qf, factor_2_qf], output_qf, mult_poly)

    return output_qf


def sbp_add(summand_1_qf, summand_2_qf, output_qf=None):
    """
    Performs addition based on the evaluation of
    `semi-boolean polynomials <https://ieeexplore.ieee.org/document/9815035>`_.

    Parameters
    ----------
    summand_1_qf : QuantumFloat
        The first summand to add.
    summand_2_qf : QuantumFloat
        The second summand to add.
    output_qf : QuantumFloat, optional
        The QuantumFloat to store the result in.
        By default, a suited new QuantumFloat is created.

    Returns
    -------
    output_qf : QuantumFloat
        A QuantumFloat containing the result of the addition.

    Examples
    --------

    We add two QuantumFloats:

    >>> from qrisp import QuantumFloat, sbp_add
    >>> qf_0 = QuantumFloat(3)
    >>> qf_1 = QuantumFloat(3)
    >>> qf_0[:] = 3
    >>> qf_1[:] = 4
    >>> qf_res = sbp_add(qf_0, qf_1)
    >>> print(qf_res)
    {7: 1.0}
    """

    if output_qf is None:
        from qrisp.qtypes.quantum_float import create_output_qf

        output_qf = create_output_qf([summand_1_qf, summand_2_qf], op="add")

    sum_poly = summand_1_qf.sb_poly(output_qf.msize) + summand_2_qf.sb_poly(
        output_qf.msize
    )

    # Apply sb encoder
    sb_polynomial_encoder([summand_1_qf, summand_2_qf], output_qf, sum_poly)

    return output_qf


def sbp_sub(summand_1_qf, summand_2_qf, output_qf=None):
    """
    Performs subtraction based on the evaluation of
    `semi-boolean polynomials <https://ieeexplore.ieee.org/document/9815035>`_.

    Parameters
    ----------
    summand_1_qf : QuantumFloat
        The QuantumFloat to subtract from.
    summand_2_qf : QuantumFloat
        The QuantumFloat to subtract.
    output_qf : QuantumFloat, optional
        The QuantumFloat to store the result in.
        By default, a suited new QuantumFloat is created.

    Returns
    -------
    output_qf : QuantumFloat
        A QuantumFloat containing the result of the subtraction.

    Examples
    --------

    We add two QuantumFloats:

    >>> from qrisp import QuantumFloat, sbp_sub
    >>> qf_0 = QuantumFloat(3)
    >>> qf_1 = QuantumFloat(3)
    >>> qf_0[:] = 3
    >>> qf_1[:] = 4
    >>> qf_res = sbp_sub(qf_0, qf_1)
    >>> print(qf_res)
    {-1: 1.0}


    """

    if output_qf is None:
        from qrisp.qtypes.quantum_float import create_output_qf

        output_qf = create_output_qf([summand_1_qf, summand_2_qf], op="sub")

    dif_poly = summand_1_qf.sb_poly(output_qf.msize) - summand_2_qf.sb_poly(
        output_qf.msize
    )

    # Apply sb encoder
    sb_polynomial_encoder([summand_1_qf, summand_2_qf], output_qf, dif_poly)

    return output_qf


# Encode the polynomial given in poly in output var
# depending on the input variables qv_list
@gate_wrap(is_qfree=True, permeability=[0])
def polynomial_encoder(qf_list, output_qf, poly, encoding_dic=None, inplace_mult=1):
    """
    Evaluates a (multivariate) sympy polynomial on a list of QuantumFloats using
    `semi-boolean polynomials <https://ieeexplore.ieee.org/document/9815035>`_.

    Parameters
    ----------
    qf_list : list[QuantumFloat]
        The list of QuantumFloats to evaluate the polynomial on.
    output_qf : QuantumFloat
        The QuantumFloat to evaluate into.
    poly : sympy expression
        The polynomial to evaluate.
    encoding_dic : dict, optional
        A dictionary which has the QuantumFloats of qf_list as keys and the associated
        sympy symbols as values. By default, the symbols of the polynomial
        will be ordered alphabetically and then matched to the order in qf_list.
    inplace_mult : int, optional
        This integer allow to perform an inplace multiplication on output_qf,
        before the polynomial is evaluated.
        Note that due to reversibility only odd numbers are supported.

    Raises
    ------
    Exception
        Provided QuantumFloat list does not include the appropriate amount of elements
        to encode given polynomial".

    Returns
    -------
    None.

    Examples
    --------

    We evaluate the polynomial $x^2 + 2y^2$ on two QuantumFloats:


    >>> from sympy import Symbol
    >>> x = Symbol("x")
    >>> y = Symbol("y")
    >>> poly = x**2 + 2*y**2
    >>> from qrisp import QuantumFloat, polynomial_encoder
    >>> x_qf = QuantumFloat(3)
    >>> y_qf = QuantumFloat(3)
    >>> x_qf[:] = 3
    >>> y_qf[:] = 2
    >>> res_qf = QuantumFloat(7)
    >>> encoding_dic = {x_qf : x, y_qf : y}
    >>> polynomial_encoder([x_qf, y_qf], res_qf, poly, encoding_dic)
    >>> print(res_qf)
    {17.0: 1.0}
    """

    if isinstance(qf_list, QuantumArray):
        qf_list = list(qf_list.flatten().qv_array)

    if encoding_dic is not None:
        symbol_list = [encoding_dic[qv.name] for qv in qf_list]

    else:
        symbol_list = get_ordered_symbol_list(poly)

    if len(symbol_list) != len(qf_list):
        raise Exception(
            "Provided QuantumFloat list does not include the appropriate amount"
            "of elements to encode given polynomial"
        )

    if not output_qf.signed:
        for qf in qf_list:
            if qf.signed:
                raise Exception(
                    "When encoding into an unsigned quantum float"
                    "provide only unsigned inputs"
                )

    sb_poly_list = [qf.sb_poly(output_qf.size) for qf in qf_list]

    repl_dic = {symbol_list[i]: sb_poly_list[i] for i in range(len(qf_list))}
    # Substitute SB-polynomials
    sb_polynomial = poly.subs(repl_dic).expand()

    # Apply sb encoder
    sb_polynomial_encoder(qf_list, output_qf, sb_polynomial, inplace_mult=inplace_mult)


def split(a, n):
    k, m = divmod(len(a), n)
    return (a[i * k + min(i, m) : (i + 1) * k + min(i + 1, m)] for i in range(n))


# Performs inplace addition on a QFT-ed variable
# by applying successive U_g gates
def U_g_inpl_adder(modified_var, summand, mult_factor=1):
    applied_phases = []
    # Extract the coefficients of the SB-polynomial of x
    from sympy import Poly

    summand_coeffs = list(Poly(summand.sb_poly(modified_var.msize)).as_dict().values())
    summand_coeffs = [float(coeff) for coeff in summand_coeffs][::-1]
    # summand_coeffs.sort()

    summand_coeffs = summand_coeffs + (summand.size - len(summand_coeffs)) * [0]

    for j in range(summand.size):
        if summand_coeffs[j] == 0:
            continue
        applied_phases.append(
            multi_controlled_U_g(
                modified_var,
                [summand[j]],
                mult_factor * summand_coeffs[j] * 2**-modified_var.exponent,
            )
        )

    return applied_phases


# This algorithm is based on the equation X*U_g(y)*X = U_g(-y)*G with G as global phase
# This is used to perform conditional addition/subtraction required for multiplication
# using algorithm 3 from https://arxiv.org/abs/2112.10537
# s = x << (n + 1)
# s− = x
# for i in range(y.size):
#   if y[i]:
#       s+= (x << i)
#   else:
#       s−= (x << i)
# return s >> 1
# We will use the function U_g_inply_adder for the inplace additions
# while performing conditional additions using CNOT gates.
# The approach is therefore a hybrid of SBP-method and traditional logic approaches
def hybrid_mult(
    x,
    y,
    output_qf=None,
    init_op="h",
    terminal_op="qft",
    phase_tolerant=False,
    cl_factor=1,
):
    """
    An advanced algorithm for multiplication which has better depth, gate-count
    and compile time than :meth:`sbp_mult <qrisp.sbp_mult>`.
    It does not support squaring a single QuantumFloat though.

    This algorithm also operates on the Fourier transform. Because of this,
    between successive multiplications targeting the same QuantumFloat
    it is not neccessary to Fourier-Transform.
    This advantage is expressed in the parameters init_op and terminal_op.
    These can be set to either 'h', 'qft' or None
    to leave out self canceling Fourier-transforms.

    Parameters
    ----------
    x : QuantumFloat
        The first factor to multiply.
    y : QuantumFloat
        The second factor to multiply.
    output_qf : QuantumFloat, optional
        The QuantumFloat to store the result in.
        By default a suited QuantumFloat is created.
    init_op : str, optional
        The operation to bring output_qf into it's Fourier-transform.
        The default is 'h'.
    terminal_op : str, optional
        The operation to bring output_qf back from it's Fourier-transform.
        The default is "qft".
    phase_tolerant : bool, optional
        If set to True, differing results introduce differing extra phases.
        This can be usefull to save resources incase this functions will get uncomputed.
        The default is False.
    cl_factor : float, optional
        Allows to multiply the result by a classical factor without any extra gates.
        The default is 1.

    Returns
    -------
    output_qf : QuantumFloat
        The QuantumFloat containing the result.

    Examples
    --------

    We multiply two QuantumFloat with eachother and an additional classical factor

    >>> from qrisp import QuantumFloat, hybrid_mult
    >>> qf_0 = QuantumFloat(3)
    >>> qf_1 = QuantumFloat(3)
    >>> qf_0[:] = 3
    >>> qf_1[:] = 4
    >>> qf_res = hybrid_mult(qf_0, qf_1, cl_factor = 2)
    >>> print(qf_res)
    {24: 1.0}

    """

    from qrisp import QFT, cx, h, merge, z

    # The two factors take asymetrical roles in this algorithm
    # This implies that there is most likely a prefered choice
    # for the roles depending on the size of the factors
    # Several trials showed that these roles work best
    if not x.size > y.size:
        x, y = y, x

    # We shift the exponent of both factors such that they are integers
    # and shift the result in the end, by the sum of both exponents
    # This allows convenient treatment of non-integer inputs
    # while only constructing the algorithm for integers
    x_exp = int(x.exponent)
    y_exp = int(y.exponent)

    x.exp_shift(-x_exp)
    y.exp_shift(-y_exp)

    # If no output_qf is given, create one
    from qrisp.qtypes.quantum_float import create_output_qf

    if isinstance(output_qf, type(None)):
        output_qf = create_output_qf([y, x], op="mul")

    else:
        output_qf.exp_shift(-(x_exp + y_exp))

    output_qf.extend(1, position=0)

    # output_qf.exp_shift(-1)
    # Since the result is right shifted at the end, this implies that the zero-th qubit
    # of the output will not contain any information (otherwise this would imply,
    # that integer multiplications can yield non-integer results).
    # However the result still needs to be able to
    # display every possible result. This is why it needs "extra working-space"
    # We therefore extend it by one qubit

    # Merge the sessions of all involved variables
    merge([output_qf, x, y])

    # Perform initial operation
    # (check the general documentation of SBP arithmetic for more details)
    if init_op == "h":
        h(output_qf)
    elif init_op == "qft":
        QFT(output_qf, exec_swap=False)
    else:
        h(output_qf[0])

    # We treat the case that y is signed by applying a negation on output_qf that is
    # conditioned on the sign qubit of y (see command [2]). This will be reversed
    # at the end of the algorithm. This implies that every addition that happens
    # in between is actually a subtraction.
    # So this acts as a sign reversal of the result.
    # We then need to make sure we multiply with the absolute of y
    # We do this by applying a bitwise negation on the mantissa of y conditioned on
    # the sign qubit (command [4]). This bitwise negation acts as
    # NOT y = -y + 1
    # This implies we need to correct the additional +1.
    # This additional +1 is equivalent to an extra +x in the result
    # Therefore we need to subtract this +x from output_qf.
    # Note that this subtraction also has to be performed
    # conditioned on the sign qubit of y.
    # Instead of adding an additional control qubit, we again follow the strategy
    # of turning a conditional subtraction into a conditional subtraction/addition
    if y.signed:
        # This performs the first part of the conditional subtraction/addition
        # for j in range(x.size):
        # multi_controlled_U_g(output_qf, [x[j]], -x_coeffs[j])
        U_g_inpl_adder(output_qf, x, -1 * cl_factor)  # command [1]

        # Negate output_qf conditioned on the sign qubit of y
        cx(y[-1], output_qf)  # command [2]

        # This would perform the second part of the subtraction/addition
        # However this command is merged into command [5] for more performance gains.
        # If commented in, this command reverses command [1]
        # if output_qf has not been negated IF the sign qubit
        # has not negated output_qf. Otherwise the subtraction of x is completed.
        # for j in range(x.size):
        # multi_controlled_U_g(output_qf, [x[j]], x_coeffs[j]) #command [3]

        # Negate mantissa of y
        cx(y[-1], y[:-1])  # command [4]

    # This performs the initial two lines of the initially mentioned
    # multiplication algorithm:
    # s = x << (n + 1)
    # s− = x

    # Note that the boolean y.signed adds the phase that would have been added
    # in command [3]
    # however not requiring another round of U_g gates
    applied_phases = U_g_inpl_adder(
        output_qf, x, cl_factor * (2 ** (y.msize) - 1 + y.signed)
    )

    # We now come to the loop of the multiplication algorithm

    # for i in range(y.size):
    #   if y[i]:
    #       s+= (x << i)
    #   else:
    #       s−= (x << i)

    # Negate output_qf in order to perform the conditional addition/subtraction

    if not phase_tolerant and terminal_op != "qft":
        hybrid_mult_anc = QuantumVariable(1)
        if y.signed:
            cx(y[-1], hybrid_mult_anc[0])

            for k in range(len(applied_phases)):
                cp(-applied_phases[k] * 2, hybrid_mult_anc[0], x[k])

    cx(y[0], output_qf)
    for i in range(y.msize):
        # Perform in-place addition
        # (note that if y[i] is active an addition has to be performed).
        # Therefore, we need to perform a subtraction here,
        # because if y[i] is active output_qf has been negated
        applied_phases = U_g_inpl_adder(output_qf, x, -(2**i) * cl_factor)

        if not phase_tolerant and terminal_op != "qft":
            if i != 0:
                cx(y[i - 1], hybrid_mult_anc[0])
            cx(y[i], hybrid_mult_anc[0])

            for k in range(len(applied_phases)):
                cp(-applied_phases[k] * 2, hybrid_mult_anc[0], x[k])
                # sbp+= -2*applied_phases[k]*Symbol("y_" + str(i))*Symbol("x_" + str(k))

        # This command is equivalent to
        # cx(y[i], output_qf)
        # cx(y[i+1], output_qf)
        # ie. performs the output_qf negation but requires less cnot gates
        if i != y.msize - 1:
            cx(y[i], y[i + 1])
            cx(y[i + 1], output_qf)
            cx(y[i], y[i + 1])
        else:
            if not y.signed:
                # This command performs the final negation
                # if y is signed, we can perform the negation
                # together with the negation conditioned on the sign qubit
                # of y (command [6])
                cx(y[i], output_qf)  # command [5]

    if not phase_tolerant and terminal_op != "qft":
        cx(y[i], hybrid_mult_anc[0])
        if y.signed:
            cx(y[-1], hybrid_mult_anc[0])

        # from sympy import Symbol, simplify
        # temp_x = sum([2**i*Symbol("x_" + str(i)) for i in range(3)], 0)
        # temp_y = sum([2**i*Symbol("y_" + str(i)) for i in range(3)], 0)

        # sbp = sbp/(2*np.pi)

        # sbp = simplify((sbp + (temp_x*temp_y).expand()/2**7))
        # print(sbp)
        hybrid_mult_anc.delete()

    if y.signed:
        # This makes sure the negation from command [5] is performed
        cx(y[i], y[-1])

        cx(y[-1], output_qf)  # command[6]

        cx(y[i], y[-1])

        # Reverse command [4]
        cx(y[-1], y[:-1])

    # Free up the qubit which we identified as containing no information

    h(output_qf[0])
    output_qf.reduce(output_qf[0])

    # Perform terminal qft
    if terminal_op == "qft":
        QFT(output_qf, inv=True, exec_swap=False)

    if not phase_tolerant and terminal_op == "qft":
        # This is a phase tolerant correction using only single qubit gates.
        # We found it by looking at the sbp of the correction using cp gates and
        # identified the sbp of x*y in it.
        for i in range(output_qf.size):
            p(-(2 * np.pi) * 2**i / 2 ** (output_qf.size + 1), output_qf[i])

        if output_qf.signed:
            z(output_qf[-1])

    output_qf.exp_shift((x_exp + y_exp))

    # Perform exponent shifts

    x.exp_shift(x_exp)
    y.exp_shift(y_exp)

    return output_qf


# Wrapper for choosing the best multiplication algorithm
def q_mult(factor_1, factor_2, target=None, method="auto"):
    if method == "auto":
        if factor_1.reg == factor_2.reg:
            return q_mult(factor_1, factor_2, target, method="sbp")
        else:
            return q_mult(factor_1, factor_2, target, method="hybrid")

    elif method == "sbp":
        from qrisp.qtypes.quantum_float import create_output_qf

        if target is None:
            target = create_output_qf([factor_1, factor_2], op="mul")

        sbp_mult(factor_1, factor_2, target)
        return target

    elif method == "hybrid":
        from qrisp.arithmetic import hybrid_mult

        return hybrid_mult(factor_1, factor_2)


def QFT_inpl_mult(qv, inplace_mult=1):
    from qrisp.misc import is_inv

    qv = list(qv)
    qv = qv[::-1]
    n = len(qv)

    if not is_inv(inplace_mult, n):
        raise Exception(
            "Tried to perform non-invertible inplace multiplication during Fourier-Transform"
        )

    # Perform QFT with inplace multiplication
    for i in range(n):
        if i != n - 1:
            h(qv[i])

        if i == n - 1:
            break

        for k in range(n - i - 1):
            if k + i + 1 != n - 1:
                cp(inplace_mult * 2 * np.pi / 2 ** (k + 2), qv[k + i + 1], qv[i])
            else:
                # The -1 here cancels some cp gates of the inverse qft
                cp((inplace_mult - 1) * 2 * np.pi / 2 ** (k + 2), qv[k + i + 1], qv[i])

    # Perform reversed QFT without inplace multiplication and without canceled steps
    for i in range(n)[::-1]:
        for k in range(n - i - 1)[::-1]:
            if k + i + 1 != n - 1:
                cp(-2 * np.pi / 2 ** (k + 2), qv[k + i + 1], qv[i])

        if i != n - 1:
            h(qv[i])

    return qv


def inpl_mult(qf, mult_int, treat_overflow=True):
    """
    Performs inplace multiplication of a :ref:`QuantumFloat` with a classical integer.
    To prevent overflow errors, this function automatically adjusts the mantissa size.
    If you want to prevent this behavior, set ``treat_overflow = False``.

    Parameters
    ----------
    qf : QuantumFloat
        The QuantumFloat to inplace-multiply.
    mult_int : int
        The integer to perform the multiplication with.
    treat_overflow : bool
        If set to ``False``, the mantissa will not be extended to prevent overflow errors. The default is ``True``.

    Examples
    --------

    We create a QuantumFloat, bring it to superposition and perform an inplace multiplication.

    >>> from qrisp import QuantumFloat, h, inpl_mult
    >>> a = QuantumFloat(5, signed = True)
    >>> h(a[0])
    >>> h(a[-1])
    >>> print(a)
    {0: 0.25, 1: 0.25, -32: 0.25, -31: 0.25}
    >>> inpl_mult(a, -5)
    >>> print(a)
    {0: 0.25, 155: 0.25, 160: 0.25, -5: 0.25}

    """

    if mult_int < 0 and not qf.signed:
        raise Exception(
            "Tried to inplace-multiply unsigned QuantumFloat with negative factor"
        )

    bit_shift = 0
    
    if int(mult_int) != mult_int:
        
        c = abs(mult_int)
        
        for i in range(32):
            if int(2**i*c) == 2**i*c:
                break
        else:
            raise Exception("Tried to inplace multiply with number of to much precision")
        
        bit_shift = -i
        mult_int = 2**i*mult_int
        
    else:
        while not mult_int % 2:
            bit_shift += 1
            mult_int = mult_int // 2

    if mult_int != 1:
        if treat_overflow:
            extension_size = int(np.ceil(np.log2(abs(mult_int))))
            qf.extend(extension_size, position=qf.size - 1)
            if qf.signed:
                cx(qf[-1], qf[-1 - extension_size : -1])

        from qrisp.arithmetic.SBP_arithmetic import QFT_inpl_mult

        QFT_inpl_mult(qf, inplace_mult=mult_int)
    
    quantum_bit_shift(qf, bit_shift, treat_overflow)
    
    if treat_overflow and bit_shift<0 and qf.signed:
        cx(qf[-1], qf[bit_shift-1:-1])
    

    
def quantum_bit_shift(qf, bit_shift, treat_overflow = True):
    
    from qrisp import cyclic_shift, control, QuantumFloat
    
    if isinstance(bit_shift, QuantumFloat):
        
        if bit_shift.signed or qf.signed:
            raise Exception("Quantum-quantum bitshifting is currently only supported for unsigned arguments")
        
        for i in range(*bit_shift.mshape):
            with control(bit_shift.significant(i)):
                quantum_bit_shift(qf, 2**i)
        
        return
    
    if treat_overflow:
        
        if bit_shift > 0:
            
            if qf.signed:
                qf.extend(bit_shift, position=qf.size-1)
            else:
                qf.extend(bit_shift, position=qf.size)
                
        else:
            qf.extend(abs(bit_shift), position=0)
            qf.exp_shift(bit_shift)
            
    if qf.signed:
        cyclic_shift(qf[:-1], bit_shift)
    else:
        cyclic_shift(qf, bit_shift)
        
        
        



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


from qrisp.misc import bin_rep
from qrisp.circuit.standard_operations import op_list


def convert_circuit(qc, target_api="thrift", transpile=True):
    if transpile or target_api == "open_api":
        from qrisp.circuit import transpile

        qc = transpile(qc)

    if target_api == "thrift":
        from qrisp.interface.thrift_interface.interface_types import (
            PortableClbit,
            PortableInstruction,
            PortableOperation,
            PortableQuantumCircuit,
            PortableQubit,
        )
    elif target_api == "open_api":
        from qrisp.interface.openapi_interface.interface_types import (
            PortableClbit,
            PortableInstruction,
            PortableOperation,
            PortableQuantumCircuit,
            PortableQubit,
        )
    elif target_api == "qrisp":
        return convert_to_qrisp(qc)
    elif target_api == "qiskit":
        return convert_to_qiskit(qc, transpile)

    circuit_qubits = [PortableQubit(identifier=qb.identifier) for qb in qc.qubits]
    circuit_clbits = [PortableClbit(identifier=cb.identifier) for cb in qc.clbits]

    res_qc = PortableQuantumCircuit(
        qubits=circuit_qubits, clbits=circuit_clbits, data=[]
    )

    for i in range(len(qc.data)):
        instr = qc.data[i]
        if instr.op.name in ["qb_alloc", "qb_dealloc"]:
            # continue
            pass

        params = []
        for j in range(len(instr.op.params)):
            value = float(instr.op.params[j])
            params.append(value)

        if instr.op.definition is not None:
            definition = convert_circuit(
                instr.op.definition, target_api, transpile=False
            )
            op = PortableOperation(
                name=instr.op.name,
                num_qubits=instr.op.num_qubits,
                num_clbits=instr.op.num_clbits,
                params=params,
                definition=definition,
            )
        else:
            op = PortableOperation(
                name=instr.op.name,
                num_qubits=instr.op.num_qubits,
                num_clbits=instr.op.num_clbits,
                params=params,
            )

        qubits = [PortableQubit(identifier=qb.identifier) for qb in instr.qubits]
        clbits = [PortableClbit(identifier=cb.identifier) for cb in instr.clbits]

        res_qc.data.append(PortableInstruction(op=op, qubits=qubits, clbits=clbits))

    return res_qc


def convert_to_qrisp(qc):
    from qrisp.circuit import (
        Clbit,
        Instruction,
        Operation,
        QuantumCircuit,
        Qubit,
        fast_append,
        op_list,
    )

    op_dict = {op().name: op for op in op_list}

    res_qc = QuantumCircuit()

    # This dic gives the qiskit qubits/clbits when presented with their identifier
    bit_dic = {}

    # Add Qubits
    for i in range(len(qc.qubits)):
        res_qc.add_qubit(Qubit(qc.qubits[i].identifier))
        bit_dic[qc.qubits[i].identifier] = res_qc.qubits[-1]

    # Add Clbits
    for i in range(len(qc.clbits)):
        res_qc.add_clbit(Clbit(qc.clbits[i].identifier))
        bit_dic[qc.clbits[i].identifier] = res_qc.clbits[-1]

    with fast_append():
        for i in range(len(qc.data)):
            instr = qc.data[i]
            # Prepare qubits
            qubit_list = [bit_dic[qubit.identifier] for qubit in instr.qubits]
            clbit_list = [bit_dic[clbit.identifier] for clbit in instr.clbits]

            if not instr.op.definition is None:
                op = convert_to_qrisp(instr.op.definition).to_op()
            elif instr.op.name in op_dict.keys():
                op = op_dict[instr.op.name](*instr.op.params)
            elif instr.op.name[5:] in op_dict.keys():
                if instr.op.name[:5] == "c_if_":
                    op = op_dict[instr.op.name[5:]](*instr.op.params).c_if()
            elif instr.op.name[-3:] == "_dg" and instr.op.name[:-3] in op_dict.keys():
                op = op_dict[instr.op.name[:-3]](*instr.op.params).inverse()
            else:
                raise Exception(
                    "Don't know how to convert operation "
                    + str(instr.op.name)
                    + " to qrisp"
                )

            res_qc.data.append(Instruction(op, qubit_list, clbit_list))

    return res_qc


# Function to convert qrisp quantum circuits to Qiskit quantum circuits
def convert_to_qiskit(qc, transpile=False):
    from qiskit import ClassicalRegister, QuantumCircuit, QuantumRegister

    if transpile:
        from qrisp.circuit.transpiler import transpile

        qc = transpile(qc)

    # Create circuit

    qiskit_qc = QuantumCircuit()

    # This dic gives the qiskit qubits/clbits when presented with their identifier
    bit_dic = {}

    # Add Qubits
    for i in range(len(qc.qubits)):
        qiskit_qc.add_register(QuantumRegister(1, name=qc.qubits[i].identifier))

        bit_dic[qc.qubits[i].identifier] = qiskit_qc.qubits[-1]

    # Add Clbits
    for i in range(len(qc.clbits)):
        qiskit_qc.add_register(ClassicalRegister(1, name=qc.clbits[i].identifier))

        bit_dic[qc.clbits[i].identifier] = qiskit_qc.clbits[-1]

    # Add Instructions
    import qiskit.circuit.library.standard_gates as qsk_gates
    from qiskit.circuit import Barrier, ControlledGate, Gate, Instruction, Parameter
    from qiskit.circuit.measure import Measure

    from qrisp.circuit import ControlledOperation
    from sympy import Symbol, lambdify, Expr
    
    
    symbol_param_dic = {}
    
    for i in range(len(qc.data)):
        op = qc.data[i].op

        params = list(op.params)
        
        qiskit_params = []
        
        for p in params:
            if isinstance(p, Expr):
                
                free_symbols = list(p.free_symbols)
                lambd_expr = lambdify(free_symbols, p)
                
                qiskit_symbs = []
                
                for s in free_symbols:
                    if not s in symbol_param_dic:
                        symbol_param_dic[s] = Parameter(str(s))
                        
                    qiskit_symbs.append(symbol_param_dic[s])
                qiskit_params.append(lambd_expr(*qiskit_symbs))
                
            else:
                qiskit_params.append(p)
        
        params = qiskit_params

        
        

        # Prepare qiskit qubits
        qubit_list = [bit_dic[qubit.identifier] for qubit in qc.data[i].qubits]
        clbit_list = [bit_dic[clbit.identifier] for clbit in qc.data[i].clbits]


        if op.name in ["qb_alloc", "qb_dealloc"]:
            continue
            temp = QuantumCircuit(1)
            qiskit_ins = temp.to_gate()
            qiskit_ins.name = op.name

        elif op.name == "gphase":
            temp_qc = QuantumCircuit(1)
            qiskit_ins = temp_qc.to_gate()
            qiskit_ins.name = "gphase"

        elif op.name == "cx":
            if hasattr(op, "ctrl_state"):
                qiskit_ins = qsk_gates.XGate().control(ctrl_state=op.ctrl_state)
            else:
                qiskit_ins = qsk_gates.XGate().control()
        elif op.name == "cz":
            if hasattr(op, "ctrl_state"):
                qiskit_ins = qsk_gates.ZGate().control(ctrl_state=op.ctrl_state)
            else:
                qiskit_ins = qsk_gates.ZGate().control()
        elif op.name == "cy":
            if hasattr(op, "ctrl_state"):
                qiskit_ins = qsk_gates.YGate().control(ctrl_state=op.ctrl_state)
            else:
                qiskit_ins = qsk_gates.YGate().control()

        elif issubclass(op.__class__, ControlledOperation):
            base_name = op.base_operation.name

            if len(base_name) == 1:
                base_name = base_name.upper()

            if op.base_operation.definition:
                qiskit_definition = convert_to_qiskit(op.base_operation.definition)
                base_gate = qiskit_definition.to_gate()
                base_gate.name = op.base_operation.name
                qiskit_ins = base_gate.control(
                    len(op.controls), ctrl_state=op.ctrl_state[::-1]
                )

            else:
                base_gate = create_qiskit_instruction(op.base_operation, params)
                qiskit_ins = base_gate.control(
                    len(op.controls), ctrl_state=op.ctrl_state[::-1]
                )

        else:
            qiskit_ins = create_qiskit_instruction(op, params)

        # Append to qiskit circuit
        qiskit_qc.append(qiskit_ins, qubit_list, clbit_list)

    # Return result
    return qiskit_qc


def create_qiskit_instruction(op, params=[]):
    import qiskit.circuit.library.standard_gates as qsk_gates
    from qiskit.circuit import Barrier
    from qiskit.circuit.measure import Measure
    
    

    if op.name == "rxx":
        qiskit_ins = qsk_gates.RXXGate(*params)
    elif op.name == "rzz":
        qiskit_ins = qsk_gates.RZZGate(*params)
    elif op.name == "ryy":
        qiskit_ins = qsk_gates.RYYGate(*params)
    
    elif op.name == "measure":
        qiskit_ins = Measure()

    elif op.name == "swap":
        qiskit_ins = qsk_gates.SwapGate()
    elif op.name == "barrier":
        qiskit_ins = Barrier(op.num_qubits)
    elif op.name == "cp":
        qiskit_ins = qsk_gates.CPhaseGate(params[0])
    elif op.name == "x":
        qiskit_ins = qsk_gates.XGate()
    elif op.name == "y":
        qiskit_ins = qsk_gates.YGate()
    elif op.name == "z":
        qiskit_ins = qsk_gates.ZGate()
    elif op.name == "rx":
        qiskit_ins = qsk_gates.RXGate(params[0])
    elif op.name == "ry":
        qiskit_ins = qsk_gates.RYGate(params[0])
    elif op.name == "rz":
        qiskit_ins = qsk_gates.RZGate(params[0])
    elif op.name == "sx":
        qiskit_ins = qsk_gates.SXGate()
    elif op.name == "h":
        qiskit_ins = qsk_gates.HGate()

    elif op.name == "p":
        qiskit_ins = qsk_gates.PhaseGate(params[0])
    elif op.name == "u1":
        qiskit_ins = qsk_gates.PhaseGate(params[0])
    elif op.name == "s":
        qiskit_ins = qsk_gates.SGate()
    elif op.name == "s_dg":
        qiskit_ins = qsk_gates.SGate().inverse()
    elif op.name == "sx":
        qiskit_ins = qsk_gates.SXGate()
    elif op.name == "sx_dg":
        qiskit_ins = qsk_gates.SXdgGate()
    elif op.name == "rzz":
        qiskit_ins = qsk_gates.RZZGate(params[0])
    elif op.name == "xxyy":
        qiskit_ins = qsk_gates.XXPlusYYGate(*params, label='(XX+YY)')
    elif op.name == "t":
        qiskit_ins = qsk_gates.TGate()
    elif op.name == "t_dg":
        qiskit_ins = qsk_gates.TGate().inverse()
    elif op.name == "u3":
        qiskit_ins = qsk_gates.U3Gate(*params)
    elif op.name == "id":
        qiskit_ins = qsk_gates.IGate()
    elif op.definition:
        qiskit_definition = convert_to_qiskit(op.definition)
        if len(op.definition.clbits):
            qiskit_ins = qiskit_definition.to_instruction()
        else:
            qiskit_ins = qiskit_definition.to_gate()
        qiskit_ins.name = op.name
    else:
        raise Exception("Could not convert operation " + str(op.name) + " to Qiskit")

    return qiskit_ins


op_dic = {op().name: op for op in op_list}
op_dic["u"] = op_dic["u3"]


def convert_from_qiskit(qiskit_qc):
    from qiskit import transpile
    from qiskit.circuit import ControlledGate, ParameterExpression, Parameter

    from qrisp import Clbit, ControlledOperation, QuantumCircuit, Qubit, op_list
    from sympy import Symbol, lambdify, Expr
    

    qc = QuantumCircuit()

    for i in range(qiskit_qc.num_qubits):
        q_reg = qiskit_qc.qubits[i]._register

        qubit_name = q_reg.name
        if q_reg.size > 1:
            qubit_name += "." + str(q_reg._bits.index(qiskit_qc.qubits[i]))

        qc.add_qubit(Qubit(qubit_name))

    for i in range(qiskit_qc.num_clbits):
        cl_reg = qiskit_qc.clbits[i]._register

        clbit_name = cl_reg.name
        if cl_reg.size > 1:
            clbit_name += "." + str(cl_reg._bits.index(qiskit_qc.clbits[i]))

        qc.add_clbit(Clbit(clbit_name))

    qb_dic = {qiskit_qc.qubits[i]: qc.qubits[i] for i in range(qiskit_qc.num_qubits)}
    cb_dic = {qiskit_qc.clbits[i]: qc.clbits[i] for i in range(qiskit_qc.num_clbits)}

    symbol_param_dic = {}
    
    from sympy import sympify
    
    for i in range(len(qiskit_qc.data)):
        qiskit_op = qiskit_qc.data[i][0]

        params = list(qiskit_op.params)
        
        qrisp_params = []

        for p in params:
            if isinstance(p, ParameterExpression):
                
                lambd_expr = sympify(ParameterExpression.sympify(p))
                
                qrisp_params.append(lambd_expr)
                
            elif isinstance(p, (float,int)):
                qrisp_params.append(float(p))
            else:
                raise Exception(f"Could not convert parameter type {type(p)}")

        params = qrisp_params


        if isinstance(qiskit_op, ControlledGate):
            controlled_gate = True
            qiskit_op = qiskit_op.base_gate
        else:
            controlled_gate = False



            """ for j in range(len(qiskit_op.params)):
            if not isinstance(qiskit_op.params[j], float):  
                if isinstance(qiskit_op.params[j], Expr):
                    pass    
                else:
                    params[j] = float(qiskit_op.params[j]) """

        try:
            op = op_dic[qiskit_op.name](*params)
        except KeyError:
            try:
                op = op_dic[qiskit_op.name.lower()](*params)
            except KeyError:
                if qiskit_op.definition is not None:
                    op = convert_from_qiskit(qiskit_op.definition).to_gate(
                        name=qiskit_op.name
                    )
                else:
                    raise Exception(
                        "Could not convert Qiskit operation "
                        + str(qiskit_op.name)
                        + " to Qrisp"
                    )

        if controlled_gate:
            qiskit_op = qiskit_qc.data[i][0]
            op = ControlledOperation(
                base_operation=op,
                num_ctrl_qubits=qiskit_op.num_ctrl_qubits,
                ctrl_state=bin_rep(qiskit_op.ctrl_state, qiskit_op.num_ctrl_qubits)[
                    ::-1
                ],
            )

        qubits = [qb_dic[qb] for qb in qiskit_qc.data[i][1]]
        clbits = [cb_dic[cb] for cb in qiskit_qc.data[i][2]]

        qc.append(op, qubits, clbits)

    return qc



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

# Created by ann81984 at 22.07.2022
from qrisp.quantum_network import QuantumNetwork
import random
import numpy as np
import time


def test_quantum_teleportation():
    # In this example we will demonstrate Qrisps quantum network simulator
    # by performing a teleportation

    # Create a Server - this server will keep track of the participants
    # and simulate the state of the network
    qn = QuantumNetwork()

    # Register clients
    qn.register_client("alice")
    qn.register_client("bob")

    # Create Bell-Pair function
    def share_bell_pair(qn, recipient_0, recipient_1):
        client_name = "bell_pair_distributor_" + str(random.randint(0, int(1e6)))
        qn.register_client(client_name)

        qn.request_qubits(2, client_name)

        telamon_qc = qn.get_clear_qc(client_name)

        telamon_qc.h(0)
        telamon_qc.cx(0, 1)

        qn.run(telamon_qc, client_name)

        qn.send_qubits(client_name, recipient_0, [telamon_qc.qubits[0]], "")
        qn.send_qubits(client_name, recipient_1, [telamon_qc.qubits[1]], "")

        return telamon_qc.qubits[0], telamon_qc.qubits[1]

    def teleport_qubit(qn, alice, bob, teleported_qubit):
        # Share Bell-Pair
        bell_pair_0, bell_pair_1 = share_bell_pair(qn, alice, bob)

        # Get the updated circuit for alice (now contains one half of the bell-pair)
        alice_qc = qn.get_clear_qc(alice)

        # Perform the Alice's steps of the quantum teleportation protocol
        alice_qc.cx(teleported_qubit, bell_pair_0)
        alice_qc.h(teleported_qubit)
        alice_qc.measure(bell_pair_0)
        alice_qc.measure(teleported_qubit)

        # Execute the circuit
        alice_res = qn.run(alice_qc, alice)

        # Perform Bob's steps of the quantum teleportation protocol
        bob_qc = qn.get_clear_qc(bob)

        # The information about the outcome alice_res is transfered via a classical channel
        if list(alice_res.keys())[0][0] == "1":
            bob_qc.x(bell_pair_1)

        if list(alice_res.keys())[0][1] == "1":
            bob_qc.z(bell_pair_1)

        return bell_pair_1

    qn.request_qubits(2, "alice")

    # Get clear circuit for alice
    alice_qc = qn.get_clear_qc("alice")

    # Apply some arbitrary transformation (the resulting state is teleported to bob)
    # alice_qc.rx(np.pi/2, 0)

    qn.run(alice_qc, "alice")

    recieved_qubit = teleport_qubit(qn, "alice", "bob", alice_qc.qubits[0])

    # Check the teleported_qubit
    bob_qc = qn.get_clear_qc("bob")
    # Measure the teleported qubit
    bob_qc.measure(recieved_qubit)

    qn.run(bob_qc, "bob")

    def gen_random_bitstring(size):
        res = np.zeros(size)
        for i in range(size):
            if random.random() > 0.5:
                res[i] = 1
        return res

    n = 50

    start_time = time.time()
    qn = QuantumNetwork()

    qn.register_client("alice")
    qn.request_qubits(n, "alice")
    alice_qc = qn.get_clear_qc("alice")

    qn.register_client("bob")
    qn.request_qubits(0, "bob")
    bob_qc = qn.get_clear_qc("bob")
    a = gen_random_bitstring(n)
    b = gen_random_bitstring(n)

    for i in range(n):
        if a[i] == 1:
            alice_qc.x(i)
        if b[i] == 1:
            alice_qc.h(i)

    qn.run(alice_qc, "alice")

    tp_qubits = []
    for i in range(n):
        tp_qubit = teleport_qubit(qn, "alice", "bob", alice_qc.qubits[i])
        tp_qubits.append(tp_qubit)
        # qn.send_qubit("alice", "bob", alice_qc.qubits[i])

    bob_qc = qn.get_clear_qc("bob")

    b_prime = gen_random_bitstring(n)

    for i in range(n):
        if b_prime[i] == 1:
            bob_qc.h(tp_qubits[i])
        bob_qc.measure(tp_qubits[i])

    res = qn.run(bob_qc, "bob")
    res = list(res.keys())[0]
    res = np.array([int(c) for c in res])

    congruent_b = b == b_prime

    temp_key_bob = res[congruent_b]
    temp_key_alice = a[congruent_b]
    k = len(temp_key_bob)

    if np.all(temp_key_bob[: k // 2] == temp_key_alice[: k // 2]):
        raise Exception("Evesdropper detected!")

    shared_key = temp_key_bob[k // 2 :]

    print(shared_key)
    assert all(i == 0 or i == 1 for i in shared_key)

    print(len(shared_key))
    assert len(shared_key) < 20

    print("Transmission rate: ", len(shared_key) / (time.time() - start_time))
    assert len(shared_key) / (time.time() - start_time) < 50
    overall_qc = qn.get_overall_qc()
    print(len(overall_qc.qubits))
    assert len(overall_qc.qubits) == 150



"""
\********************************************************************************
* Copyright (c) 2023 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""


import numpy as np

from qrisp import QuantumCircuit
from qrisp.misc import gate_wrap

# Implementation of the modular Cuccaro-Adder https://arxiv.org/pdf/quant-ph/0410184.pdf
# Has complexity O(log(N)) (Fourier-Adding has O(log(N)^2))


def MAJ_gate():
    qc = QuantumCircuit(3)

    qc.cx(2, 1)
    qc.cx(2, 0)
    qc.mcx([0, 1], 2)

    result = qc.to_gate()

    result.name = "MAJ"

    return result


def UMA_gate(mode=2):
    qc = QuantumCircuit(3)
    if mode == 2:
        qc.mcx([0, 1], 2)
        qc.cx(2, 0)
        qc.cx(0, 1)

    if mode == 3:
        qc.x(1)
        qc.cx(0, 1)
        qc.mcx([0, 1], 2)
        qc.x(1)
        qc.cx(2, 0)
        qc.cx(2, 1)

    result = qc.to_gate()
    result.name = "UMA"
    return result


# Performs inplace addition on qv2, ie.
# qv1 += qv2


def cuccaro_procedure(qs, qubit_list_1, qubit_list_2, output_qubit=None):
    if len(qubit_list_1) != len(qubit_list_2):
        raise Exception(
            "Tried to call Cuccaro-procedure with qubit lists of unequal length"
        )

    from qrisp.core import QuantumVariable

    # Request ancilla Qubit
    ancilla = QuantumVariable(1)

    # Prepare MAJ/UMA gate qubits
    slot_1_qbs = ancilla.reg + qubit_list_1[:-1]
    slot_2_qbs = qubit_list_2
    slot_3_qbs = qubit_list_1

    iterations = len(slot_1_qbs)

    # Perform 1st step of the modular addition section
    for i in range(iterations):
        qbits = [slot_1_qbs[i], slot_2_qbs[i], slot_3_qbs[i]]
        qs.append(MAJ_gate(), qbits)

    # Calculate output qbit
    if output_qubit:
        qs.cx(qbits[2], output_qubit)

    # Perform UMA iterations
    for i in range(iterations - 1, -1, -1):
        qbits = [slot_1_qbs[i], slot_2_qbs[i], slot_3_qbs[i]]

        qs.append(UMA_gate(), qbits)

    # Detete ancilla
    ancilla.delete()


# Wrapper for the Cuccaro procedure
# This function mainly serves to determine the input qubits of the Cucarro procedure
# depending on the shape of the input QuantumFloats
# The effect of this function is
# qv1 += qv2


@gate_wrap(is_qfree=True, permeability=[1])
def inpl_add(
    qf1,
    qf2,
    ignore_rounding_error=False,
    ignore_overflow_error=False,
    adder="thapliyal",
):
    """
    Performs in-place addition of the second argument onto the first.
    In Python syntax: ::

        qf1 += qf2

    There are two different algorithms available:
    The `Thapliyal adder <https://arxiv.org/abs/1712.02630>`_
    and the `Cuccaro adder <https://arxiv.org/abs/quant-ph/0410184>`_

    Parameters
    ----------
    qf1 : QuantumFloat
        The QuantumFloat that will be in-place modified.
    qf2 : QuantumFloat
        The QuantumFloat that is being added.
    ignore_rounding_error : bool, optional
        If set to False, an Exception will be raised if qf2 has higher precision
        than qf1. The default is False.
    ignore_overflow_error : bool, optional
        If set to False, an Exception will be raised if qf2 has higher maximum
        significance than qf2. The default is False.
    adder : str, optional
        Specifies the adder. Available are "thapliyal" and "cuccaro".
        The default is "thapliyal".

    Raises
    ------
    Exception
        Tried to add signed QuantumFloat onto non signed QuantumFloat.

    Examples
    --------

    We create two QuantumFloats and apply the inplace adder

    >>> from qrisp import QuantumFloat, inpl_add
    >>> qf_0 = QuantumFloat(5)
    >>> qf_1 = QuantumFloat(5)
    >>> qf_0[:] = 4
    >>> qf_1[:] = 3
    >>> inpl_add(qf_0, qf_1)
    >>> print(qf_0)
    {7.0: 1.0}
    """

    qs = qf1.qs

    # If qf2 has lower exponent, the qf2 bits with less significance than all of qf1
    # can not be added to qf1 (rounding error)
    if not ignore_rounding_error and qf1.exponent > qf2.exponent:
        raise Exception(
            "Tried to add QuantumFloat to QuantumFloat of lower precision"
            " (set ignore_rounding_error = True)"
        )

    # If qf2 has higher maximum significance than qf1, the qf2 bits with higher
    # significance than all of qf1 can not be added to qf1 (overflow error)
    if not ignore_overflow_error and qf1.mshape[1] < qf2.mshape[1]:
        raise Exception(
            "Tried to add QuantumFloat to QuantumFloat of lower precision"
            " (set ignore_overflow_error = True)"
        )

    # Determine the significance range
    # For instance for a QuantumFloat with datashape [-2,3]
    # has significance range [-2,-1,0,1,2,3]
    significance_range_qf1 = list(range(qf1.mshape[0], qf1.mshape[1] + 1))
    significance_range_qf2 = list(range(qf2.mshape[0], qf2.mshape[1] + 1))

    # Determine the intersection of the significance ranges
    signficance_range_intersetion = list(
        set(significance_range_qf1).intersection(significance_range_qf2)
    )

    # Determine maximum and minimum significance of the addition
    # The maximum significance is the maximum significance of qf1
    # (since we are adding in-place)
    # The minimum significance is the minimum of the intersection because
    # if qf2.exponent < qf1.exponent
    # we are allowed to ignore possible rounding errors
    # (skip all significances lower than qf1.exponent)
    # if qf2.exponent > qf1.exponent
    # The qbits of qf1 with lower significance than all of qf2
    # stay unchanged during addition

    max_sig = max(significance_range_qf1)
    min_sig = min(signficance_range_intersetion)

    # If the maximum significance is higher than the maximum signficance of qf2, we need
    # to augment some ancilla qubits because the Cuccaro-procedure requires equal
    # amount of input qubits

    from qrisp.core import QuantumVariable

    if max_sig > max(significance_range_qf2):
        # print(max_sig-max(significance_range_qf2))
        ancilla_var = QuantumVariable(
            max_sig - max(significance_range_qf2) - int(qf2.signed)
        )
        augmented_qf2_qbs = qf2.reg + ancilla_var.reg
    else:
        augmented_qf2_qbs = qf2.reg

    # Determine the bit window of which bits should participate in the Cuccaro-procedure
    bit_window_1 = [
        significance_range_qf1.index(min_sig),
        significance_range_qf1.index(max_sig),
    ]

    if max_sig > max(significance_range_qf2):
        bit_window_2 = [significance_range_qf2.index(min_sig), len(augmented_qf2_qbs)]
    else:
        bit_window_2 = [
            significance_range_qf2.index(min_sig),
            significance_range_qf2.index(max_sig),
        ]

    # print(max_sig)
    # print(ancilla_var_.size)
    # print(significance_range_qf1)
    # print(significance_range_qf2)
    # print(bit_window_1)
    # print(bit_window_2)

    # Determine qubit lists
    qubit_list_1 = qf1.reg[bit_window_1[0] : bit_window_1[1]]
    qubit_list_2 = augmented_qf2_qbs[bit_window_2[0] : bit_window_2[1]]

    # print(len(qubit_list_1))
    # print(len(qubit_list_2))

    # Now we treat the sign bits
    if qf2.signed and not qf1.signed:
        raise Exception("Tried to add signed QuantumFloat to non signed QuantumFloat")

    # The signs are handled via modular arithmetic
    # According to the Cuccaro-Paper this is done via manipulating
    # the final qubits of the input list
    # if True:
    if qf1.signed:
        qubit_list_1.append(qf1[-1])

        if qf2.signed:
            qubit_list_2.append(qf2[-1])
        else:
            ancilla_var_2 = QuantumVariable(1)
            qubit_list_2.append(ancilla_var_2[0])

    # If qf2 is signed and we augmented Qubits, we need to make sure
    # the augmented qubits are set to 1 if qf2 is in a negative state
    # (for details on this check how negative numbers are encoded)
    if qf2.signed and max_sig > max(significance_range_qf2):
        for i in range(ancilla_var.size):
            qs.cx(qf2[-1], ancilla_var[i])

    if adder == "cuccaro":
        cuccaro_procedure(
            qs, qubit_list_2[:-1], qubit_list_1[:-1], output_qubit=qubit_list_1[-1]
        )
    elif adder == "thapliyal":
        thapliyal_procedure(
            qs, qubit_list_2[:-1], qubit_list_1[:-1], output_qubit=qubit_list_1[-1]
        )
    else:
        raise Exception("Adder " + adder + " not implemented")

    # Perform 2nd step of the modular addition section
    # qs.cx(qubit_list_2[-2], qubit_list_1[-1])
    qs.cx(qubit_list_2[-1], qubit_list_1[-1])

    # Remove the 1s from the augmented qubits and delete
    if len(augmented_qf2_qbs) != len(qf2.reg):
        if qf2.signed:
            for i in range(ancilla_var.size):
                qs.cx(qf2[-1], ancilla_var[i])

    try:
        ancilla_var.delete()
    except NameError:
        pass

    if qf1.signed and not qf2.signed:
        ancilla_var_2.delete()


# Workaround for protecting the docstring from the decorator
temp = inpl_add.__doc__
inpl_add = gate_wrap(inpl_add)
inpl_add.__doc__ = temp


# Adder based on https://arxiv.org/abs/1712.02630


def TR_gate():
    qc = QuantumCircuit(3)
    qc.crx(-np.pi / 2, 1, 2)
    qc.p(-np.pi / 4, 1)
    qc.cx(0, 1)

    qc.p(np.pi / 4, 0)
    qc.crx(np.pi / 2, 0, 2)

    qc.p(np.pi / 4, 1)
    qc.crx(np.pi / 2, 1, 2)

    # Error in Thapliyal paper? Doesnt work if there is no inverse here
    result = qc.to_gate().inverse()
    result.name = "TR"
    return result


def thapliyal_procedure(qc, qubit_list_1, qubit_list_2, output_qubit):
    if len(qubit_list_1) != len(qubit_list_2):
        raise Exception(
            "Tried to call Thapliyal-procedure with qubit lists of unequal length"
        )

    n = len(qubit_list_1)

    # Step 1
    for i in range(1, n):
        qc.cx(qubit_list_1[i], qubit_list_2[i])

    # Step 2
    qc.cx(qubit_list_1[-1], output_qubit)

    for i in range(n - 2, 0, -1):
        qc.cx(qubit_list_1[i], qubit_list_1[i + 1])

    # Step 3
    for i in range(n - 1):
        qc.mcx([qubit_list_1[i], qubit_list_2[i]], qubit_list_1[i + 1])

    # Step 4
    qc.append(TR_gate(), [qubit_list_1[-1], qubit_list_2[-1], output_qubit])

    for i in range(n - 2, -1, -1):
        qc.append(TR_gate(), [qubit_list_1[i], qubit_list_2[i], qubit_list_1[i + 1]])

    # Step 5
    for i in range(1, n - 1):
        qc.cx(qubit_list_1[i], qubit_list_1[i + 1])

    # Step 6
    for i in range(1, n):
        qc.cx(qubit_list_1[i], qubit_list_2[i])
